/*---------------------------------------------------------------------------*\
 *   
 *   Copyright (C) 2006 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
\*---------------------------------------------------------------------------*/

#include <stdarg.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/if_vlan.h>
#include <linux/avm_cpmac.h>
#include <linux/in.h>
#include <linux/ip.h>
#include "cpmac_sched.h"
#include "cpmac_debug.h"
#include "cpmac_eth.h"


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static inline void destroy_mc_list(struct dev_mc_list *mc_list)
{
	struct dev_mc_list *dmi = mc_list;
	struct dev_mc_list *next;

	while(dmi) {
		next = dmi->next;
		kfree(dmi);
		dmi = next;
	}
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void copy_mc_list(struct dev_mc_list *mc_list,
                         struct cpmac_devinfo *devinfo)
{
	struct dev_mc_list *dmi, *new_dmi;

	destroy_mc_list(devinfo->old_mc_list);
	devinfo->old_mc_list = 0;

	for (dmi = mc_list; dmi != NULL; dmi = dmi->next) {
		new_dmi = kmalloc(sizeof(*new_dmi), GFP_ATOMIC);
		if (new_dmi == NULL) {
			DEB_ERR("cpmac: cannot allocate memory. "
			       "Multicast may not work properly from now.\n");
			return;
		}

		/* Copy whole structure, then make new 'next' pointer */
		*new_dmi = *dmi;
		new_dmi->next = devinfo->old_mc_list;
		devinfo->old_mc_list = new_dmi;
	}
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void flush_mc_list(struct net_device *dev)
{
    struct cpmac_devinfo *devinfo = ((struct cpmac_devinfo *)(dev->priv));
	struct dev_mc_list *dmi = dev->mc_list;

	while (dmi) {
		DEB_TRC("%s: del %.2x:%.2x:%.2x:%.2x:%.2x:%.2x mcast address from vlan interface\n",
		       dev->name,
		       dmi->dmi_addr[0],
		       dmi->dmi_addr[1],
		       dmi->dmi_addr[2],
		       dmi->dmi_addr[3],
		       dmi->dmi_addr[4],
		       dmi->dmi_addr[5]);
		dev_mc_delete(dev, dmi->dmi_addr, dmi->dmi_addrlen, 0);
		dmi = dev->mc_list;
	}

	/* dev->mc_list is NULL by the time we get here. */
	destroy_mc_list(devinfo->old_mc_list);
	devinfo->old_mc_list = NULL;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static inline int dmi_equals(struct dev_mc_list *dmi1,
                             struct dev_mc_list *dmi2)
{
	return ((dmi1->dmi_addrlen == dmi2->dmi_addrlen) &&
		(memcmp(dmi1->dmi_addr, dmi2->dmi_addr, dmi1->dmi_addrlen) == 0));
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int should_add_mc(struct dev_mc_list *dmi, struct dev_mc_list *mc_list)
{
	struct dev_mc_list *idmi;

	for (idmi = mc_list; idmi != NULL; ) {
		if (dmi_equals(dmi, idmi)) {
			if (dmi->dmi_users > idmi->dmi_users)
				return 1;
			else
				return 0;
		} else {
			idmi = idmi->next;
		}
	}

	return 1;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void cpmaceth_set_multicast_list(struct net_device *dev)
{
    struct cpmac_devinfo *devinfo = ((struct cpmac_devinfo *)(dev->priv));
	struct dev_mc_list *dmi;
	struct net_device *real_dev;
	int inc;

	real_dev = devinfo->real_dev;

	inc = dev->promiscuity - devinfo->old_promiscuity;
	if (inc) {
		DEB_INFO("%s: dev_set_promiscuity(%s, %d)\n",
			   dev->name, real_dev->name, inc);
		dev_set_promiscuity(real_dev, inc); /* found in dev.c */
		devinfo->old_promiscuity = dev->promiscuity;
	}

	inc = dev->allmulti - devinfo->old_allmulti;
	if (inc) {
		DEB_INFO("%s: dev_set_allmulti(%s, %d)\n",
			   dev->name, real_dev->name, inc);
		dev_set_allmulti(real_dev, inc); /* dev.c */
		devinfo->old_allmulti = dev->allmulti;
	}

	/* looking for addresses to add to master's list */
	for (dmi = dev->mc_list; dmi != NULL; dmi = dmi->next) {
		if (should_add_mc(dmi, devinfo->old_mc_list)) {
			dev_mc_add(real_dev, dmi->dmi_addr, dmi->dmi_addrlen, 0);
			DEB_TRC("%s: add %.2x:%.2x:%.2x:%.2x:%.2x:%.2x mcast address to interface %s\n",
				   dev->name,
				   dmi->dmi_addr[0],
				   dmi->dmi_addr[1],
				   dmi->dmi_addr[2],
				   dmi->dmi_addr[3],
				   dmi->dmi_addr[4],
				   dmi->dmi_addr[5],
				   real_dev->name);
		}
	}

	/* looking for addresses to delete from master's list */
	for (dmi = devinfo->old_mc_list; dmi != NULL; dmi = dmi->next) {
		if (should_add_mc(dmi, dev->mc_list)) {
			/* if we think we should add it to the new list, then we should really
			 * delete it from the real list on the underlying device.
			 */
			dev_mc_delete(real_dev, dmi->dmi_addr, dmi->dmi_addrlen, 0);
			DEB_TRC("%s: del %.2x:%.2x:%.2x:%.2x:%.2x:%.2x mcast address from interface %s\n",
				   dev->name,
				   dmi->dmi_addr[0],
				   dmi->dmi_addr[1],
				   dmi->dmi_addr[2],
				   dmi->dmi_addr[3],
				   dmi->dmi_addr[4],
				   dmi->dmi_addr[5],
				   real_dev->name);
		}
	}

	/* save multicast list */
	copy_mc_list(dev->mc_list, devinfo);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static struct net_device_stats *
cpmaceth_get_stats(struct net_device *dev)
{
   struct cpmac_devinfo *devinfo = ((struct cpmac_devinfo *)(dev->priv));
   return &devinfo->stats;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int cpmaceth_change_mtu(struct net_device *dev, int new_mtu)
{
   struct cpmac_devinfo *devinfo = ((struct cpmac_devinfo *)(dev->priv));
   if (devinfo->real_dev->mtu < new_mtu)
	  return -ERANGE;
   dev->mtu = new_mtu;
   return 0;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static const u8 _prio2band[TC_PRIO_MAX+1] =
	{ 1, 2, 2, 2, 1, 2, 0, 0 , 1, 1, 1, 1, 1, 1, 1, 1 };

static inline u8 prio2band(u32 prio)
{
   return _prio2band[prio&TC_PRIO_MAX];
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int
cpmaceth_hard_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
   struct cpmac_devinfo *devinfo = ((struct cpmac_devinfo *)(dev->priv));
   struct net_device_stats *stats = &devinfo->stats;
   struct vlan_ethhdr *veth = (struct vlan_ethhdr *)(skb->data);
   struct iphdr *iph;

   if (veth->h_vlan_proto != __constant_htons(ETH_P_8021Q)) {
      skb = __vlan_put_tag(skb, devinfo->VID);
      if (!skb) {
         stats->tx_dropped++;
         return NETDEV_TX_OK;
      }
   }

   veth = (struct vlan_ethhdr *)(skb->data);
   if (   veth->h_vlan_encapsulated_proto == __constant_htons(ETH_P_IP)
       && (iph = (struct iphdr *)(veth+1)) != 0) {
	  if (MULTICAST(iph->daddr)) {
         skb->priority = CPMAC_PRIO_ETH_MULTICAST;
	  } else if (iph->protocol == IPPROTO_TCP && ntohs(iph->tot_len) <= 64) {
         skb->priority = CPMAC_PRIO_ETH_ACK;
	  } else {
         skb->priority = CPMAC_PRIO_ETH_0+prio2band(skb->priority);
	  }
   } else {
      skb->priority = CPMAC_PRIO_ETH_0+prio2band(skb->priority);
   }
   /* use queue 0 */
   skb->uniq_id &= 0xffffff;

   stats->tx_packets++;
   stats->tx_bytes += skb->len;
   skb->dev = devinfo->real_dev;
   dev->trans_start = jiffies;
   dev_queue_xmit(skb);
   return NETDEV_TX_OK;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static inline int is_tcpack(struct sk_buff *skb)
{
   return    skb->nh.iph
          && skb->nh.iph->protocol == IPPROTO_TCP
	      && ntohs(skb->nh.iph->tot_len) <= 64;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int
cpmacwan_hard_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
   struct cpmac_devinfo *devinfo = ((struct cpmac_devinfo *)(dev->priv));
   struct net_device_stats *stats = &devinfo->stats;
   struct vlan_ethhdr *veth = (struct vlan_ethhdr *)(skb->data);
   int port;

   if (skb->nh.iph == 0) {
	  skb->priority = CPMAC_PRIO_WAN_PPP;
   } else if ((port = (skb->uniq_id >> 24)) != 0) { /* from cpmac */
	  if (port == 3 || port == 4) {
		 if (is_tcpack(skb))
            skb->priority = CPMAC_PRIO_WAN_HFORW_ACK;
		 else
            skb->priority = CPMAC_PRIO_WAN_HFORW_0+prio2band(skb->priority);
	  } else {
		 if (is_tcpack(skb))
            skb->priority = CPMAC_PRIO_WAN_LFORW_ACK;
		 else
            skb->priority = CPMAC_PRIO_WAN_LFORW_0+prio2band(skb->priority);
	  }
   } else if (skb->sk) { /* local system */
      if (skb->nh.iph->protocol == IPPROTO_IGMP) {
	     skb->priority = CPMAC_PRIO_WAN_IGMP;
      } else if (skb->nh.iph->protocol == IPPROTO_UDP) {
		 skb->priority = CPMAC_PRIO_WAN_VOIP;
	  } else {
	     skb->priority = CPMAC_PRIO_WAN_LOCAL_0+prio2band(skb->priority);
	  }
   } else { /* from wlan */
	  if (is_tcpack(skb))
		 skb->priority = CPMAC_PRIO_WAN_LFORW_ACK;
	  else
	     skb->priority = CPMAC_PRIO_WAN_LFORW_0+prio2band(skb->priority);
   }
   /* use queue 1 */
   skb->uniq_id &= 0xffffff;
   skb->uniq_id |= (1<<24);

   if (veth->h_vlan_proto != __constant_htons(ETH_P_8021Q)) {
      skb = __vlan_put_tag(skb, devinfo->VID);
	  if (!skb) {
		 stats->tx_dropped++;
         return NETDEV_TX_OK;
	  }
   }
   stats->tx_packets++;
   stats->tx_bytes += skb->len;
   skb->dev = devinfo->real_dev;
   dev->trans_start = jiffies;
   dev_queue_xmit(skb);
   return NETDEV_TX_OK;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int cpmaceth_open(struct net_device *dev)
{
   struct cpmac_devinfo *devinfo = ((struct cpmac_devinfo *)(dev->priv));

   if (!(devinfo->real_dev->flags & IFF_UP))
		return -ENETDOWN;

   DEB_INFO("cpmaceth_open(%s)\n", dev->name);

   netif_start_queue(dev);

   return 0;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int cpmaceth_stop(struct net_device *dev)
{
	flush_mc_list(dev);
    DEB_INFO("cpmaceth_close(%s)\n", dev->name);

    if (!netif_queue_stopped(dev))
       netif_stop_queue(dev);
	return 0;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int cpmaceth_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
    struct cpmac_devinfo *devinfo = ((struct cpmac_devinfo *)(dev->priv));
	struct net_device *real_dev = devinfo->real_dev;
	struct ifreq ifrr;
	int err = -EOPNOTSUPP;

	strncpy(ifrr.ifr_name, real_dev->name, IFNAMSIZ);
	ifrr.ifr_ifru = ifr->ifr_ifru;

	switch(cmd) {
	case SIOCGMIIPHY:
	case SIOCGMIIREG:
	case SIOCSMIIREG:
    case AVM_CPMAC_IOCTL_SET_CONFIG_N:
    case AVM_CPMAC_IOCTL_GET_CONFIG_N:
		if (real_dev->do_ioctl && netif_device_present(real_dev)) 
			err = real_dev->do_ioctl(real_dev, &ifrr, cmd);
		break;

	case SIOCETHTOOL:
		err = dev_ethtool(&ifrr);
        break;
	}

	if (!err) 
		ifr->ifr_ifru = ifrr.ifr_ifru;

	return err;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static struct net_device *
cpmac_register_device(char *name,
                      struct net_device *real_dev,
					  unsigned short VID,
					  int is_wan) {
    struct net_device *dev;
    dev = alloc_netdev(sizeof(struct cpmac_devinfo), name, ether_setup);
    if(dev) {
        struct cpmac_devinfo *devinfo;
	    int ret;

	    devinfo = ((struct cpmac_devinfo *)(dev->priv));
	    memset(devinfo, 0, sizeof(struct cpmac_devinfo));
	    devinfo->real_dev = real_dev;
	    devinfo->VID = VID;
	    devinfo->is_wan = is_wan;

        dev->flags = real_dev->flags;
	    dev->flags &= ~IFF_UP;
	    dev->mtu = real_dev->mtu;
	    dev->type = real_dev->type;
	    dev->hard_header_len = real_dev->hard_header_len + VLAN_HLEN;
        memcpy(dev->broadcast, real_dev->broadcast, real_dev->addr_len);
        memcpy(dev->dev_addr, real_dev->dev_addr, real_dev->addr_len);
        dev->addr_len = real_dev->addr_len;

	    dev->get_stats = cpmaceth_get_stats;
	    dev->tx_queue_len = 0;

	    /* set up method calls */
	    dev->change_mtu = cpmaceth_change_mtu;
	    dev->open = cpmaceth_open;
	    dev->stop = cpmaceth_stop;
	    dev->set_multicast_list = cpmaceth_set_multicast_list;
	    dev->destructor = free_netdev;
	    dev->do_ioctl = cpmaceth_ioctl;

        if(is_wan)
           dev->hard_start_xmit = cpmacwan_hard_start_xmit;
        else
           dev->hard_start_xmit = cpmaceth_hard_start_xmit;

	    ret = register_netdevice(dev);
	    if (ret != 0) {
	       DEB_ERR("cpmac: register_netdev(%s) failed rc = %d\n",
	     				   dev->name, ret);
	       free_netdev(dev);
	       dev = 0;
	    }
	}
	return dev;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
struct net_device *
cpmac_register_eth_device(char *name,
                          struct net_device *real_dev,
					      unsigned short VID)
{
	return cpmac_register_device(name, real_dev, VID, 0);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
struct net_device *
cpmac_register_wan_device(char *name,
                          struct net_device *real_dev,
					      unsigned short VID)
{
	return cpmac_register_device(name, real_dev, VID, 1);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_unregister_device(struct net_device *dev)
{
   unregister_netdevice(dev);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void cpmac_update_device(struct net_device *dev, unsigned short VID)
{
   struct cpmac_devinfo *devinfo = ((struct cpmac_devinfo *)(dev->priv));
   devinfo->VID = VID;
}

