/*---------------------------------------------------------------------------*\
 *   
 *   Copyright (C) 2006 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
\*---------------------------------------------------------------------------*/
 
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <asm/mips-boards/ohioint.h>
#include <linux/spinlock.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/list.h>
#include <linux/init.h>
#include "ahcidriver.h"
#include "ahciroothub.h"
#include "ahcidef.h"
#include "ahcireg.h"
#include "ahcidesc.h"
#include "ahciops.h"
#include "ahcilist.h"
#include "ahcirx.h"
#include "ahciurb.h"
#include "ahciirq.h"
#include "ahcimem.h"
#include "ahciport.h"

#include "hcd.h"				/* usb_operations */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
MODULE_LICENSE("GPL");
MODULE_AUTHOR("AVM GmbH");
MODULE_DESCRIPTION("AVM Host Controller Interface Driver");

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
PAHCI_REGISTER		AHCI_BaseAddress	= NULL;
static unsigned		AHCI_RegisterOffset	= 0x4000;
static unsigned		AHCI_IntLine		= 1; /* Interrupt line # */

MODULE_PARM (AHCI_BaseAddress, "i");
MODULE_PARM (AHCI_RegisterOffset, "i");
MODULE_PARM (AHCI_IntLine, "i");

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static inline int irq_mapping (int irq) {

	/* FIXME: How can kernel 2.4's LNXINTNUM(AHCI_IntLine) be translated
	 *        into the world of ohioint.h constants?
	 */
	if (irq == 1) return OHIOINT_EXT_0;	/* ... or OHIOINT_EXT_1 ? */
	assert (0);
	return -1;
} /* irq_mapping */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

/* Prototyps and replacements for the old vanishing hcd interface... */

static struct device ahci_fake_device;

extern int usb_register_bus(struct usb_bus * bus);
extern void usb_deregister_bus (struct usb_bus * bus);

static void ahci_free_bus (struct usb_bus * bus) {
	PrintInfo ("ahci_free_bus\n");
	assert (bus != NULL);
	kfree (bus);
} /* ahci_free_bus */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static void __exit deinit_ahci (void);

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

/* Forward declarations... */

static int ahci_get_current_frame_number (struct usb_device *usb_dev);
static void* ahci_usb_buffer_alloc (struct usb_bus* bus, size_t size, unsigned flags, dma_addr_t * pdma);
static void ahci_usb_buffer_free (struct usb_bus *bus, size_t size, void *addr, dma_addr_t dma);

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
struct usb_operations	ahci_device_operations	= {
	
	.get_frame_number =	ahci_get_current_frame_number,
	.submit_urb =		ahci_submit_urb,
	.unlink_urb =		ahci_unlink_urb,
 	.buffer_alloc = 	ahci_usb_buffer_alloc,
 	.buffer_free =		ahci_usb_buffer_free,
 	.disable =		ahci_endpoint_disable,
 	
#if defined (CONFIG_USB_SUSPEND)
#error CONFIG_USB_SUSPEND is currently NOT supported!
	.hub_suspend =		NULL,
	.hub_resume =		NULL,
#endif
};
static struct device_driver	ahci_driver	= {

	.name =			"ahci",
};

unsigned		AHCI_TotalMemSize	= 0;

struct ahci_struct	ahci			= {0};

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static void* ahci_usb_buffer_alloc (struct usb_bus* bus, size_t size, unsigned flags, dma_addr_t * pdma) {
	return kmalloc (size, flags);
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static void ahci_usb_buffer_free (struct usb_bus *bus, size_t size, void *addr, dma_addr_t dma) {
	kfree (addr);
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static int ahci_get_current_frame_number (struct usb_device *usb_dev) 
{
	PrintInfo ("ahci_get_current_frame_number %p called\n", usb_dev);
	return AHCI_GetFrameNumber();
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static int ahci_init_bus (void) {
	ahci.bus = usb_alloc_bus (&ahci_device_operations);

	info (ahci.bus != 0);
	if (!ahci.bus) {
		return -ENOMEM;
	}

	ahci.bus->bus_name = "ahci_hcd";
	ahci.bus->hcpriv = &ahci;
	ahci.rh_address = 0;

        device_initialize(&ahci_fake_device);
	ahci_fake_device.driver = &ahci_driver;
	kobject_set_name (&ahci_fake_device.kobj, "ahci-hcd");
	kobject_add (&ahci_fake_device.kobj);
	ahci.bus->controller = &ahci_fake_device;
	usb_register_bus (ahci.bus);

	return 0;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static void ahci_deinit_bus (void) {
	
	info (ahci.bus != NULL);
	if (ahci.bus != NULL) {
		usb_deregister_bus (ahci.bus);
		ahci_free_bus (ahci.bus);
		ahci.bus = NULL;
		kobject_del (&ahci_fake_device.kobj);
	}
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static void __exit deinit_ahci (void) {
	PrintInfo ("Shutting down AHCI...\n");
	ahci_deinit_root_hub (&ahci);
	ahci_deinit_bus ();
	free_irq (irq_mapping (AHCI_IntLine), (void *) AHCI_BaseAddress);
	tasklet_kill (&ahci.completion_tasklet);
	AHCI_MemoryDeinit ();
	PrintInfo ("AHCI shut down\n");
} /* deinit_ahci */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
int __init init_ahci (void) {
	int		res = -1;
	unsigned	Value;
	
#	if defined (CONFIG_AHCI_DEBUG)
	set_current_state (TASK_INTERRUPTIBLE);
	schedule_timeout (1 * HZ);
#	endif    

	spin_lock_init(&ahci.list_lock);
	
	INIT_LIST_HEAD (&ahci.completion_list);
	INIT_LIST_HEAD (&ahci.control_pending_list);
	//TO DO
	//INIT_LIST_HEAD (&ahci.iso_pending_list);
	INIT_LIST_HEAD (&ahci.int_pending_list);
	INIT_LIST_HEAD (&ahci.bulk_pending_list);
	
	PrintInfo ("AHCI (" __DATE__ "/" __TIME__ ")\n");
	
	PrintInfo ("Module parameters:\n");
	PrintInfo ("AHCI_BaseAddress    0x%p\n", AHCI_BaseAddress);
	PrintInfo ("AHCI_RegisterOffset 0x%0X\n", AHCI_RegisterOffset);
	PrintInfo ("AHCI_IntLine        0x%0X\n", AHCI_IntLine);
	
	if (AHCI_BaseAddress == NULL) {
		return -1;
	}
    
	AHCI_InitRegisterMapping (AHCI_RegisterOffset);
	
	Value = AHCI_READ_REGISTER (pAHCI_RevisionIDReg);
	AHCI_TotalMemSize = AHCI_RIDR_maxRamAddress (Value) + 4;

	PrintInfoRelease ("\tAHCI RevisionID = 0x%02X, RamSize = %u, NumPorts= %u\n", 
			AHCI_RIDR_revisionID(Value),
			AHCI_TotalMemSize, 
			AHCI_RIDR_numPorts(Value) + 1);
		
    	res = AHCI_ResetHost ();
    	if (res == 0) {
		AHCI_ConfigRxMemorySize	(CONFIG_AHCI_RX_MEMSIZE);

		res = AHCI_MemoryInit (CONFIG_AHCI_RX_MEMSIZE, AHCI_TotalMemSize - CONFIG_AHCI_RX_MEMSIZE);
	}
	if (res == 0) {
		ahci_InitPorts (AHCI_RIDR_numPorts(Value));
		tasklet_init (&ahci.completion_tasklet, urb_completion_func, 0);
		res = request_irq (
			irq_mapping (AHCI_IntLine), 
			ahci_irq, 
			SA_INTERRUPT | SA_SHIRQ, 
			"ahci-hcd", 
			(void *) AHCI_BaseAddress
		);
		if (res != 0) {
			PrintError ("Could not register AHCI IRQ #%X, error: %d\n", irq_mapping (AHCI_IntLine), res);
			res = -EBUSY;
		} else {
			PrintInfo ("AHCI IRQ 0x%02x registered\n", irq_mapping (AHCI_IntLine));
			/* 0 ms IRQ delay */
			AHCI_WRITE_REGISTER (pAHCI_InterruptConfigReg, 0);
			AHCI_WRITE_REGISTER (pAHCI_InterruptStatusReg, 0xF);

			AHCI_WRITE_REGISTER (pAHCI_InterruptEnableReg, 
				(AHCI_IER_newStatusWord_FLAG | AHCI_IER_portStatusChange_FLAG | AHCI_IER_hostHalted_FLAG));
		}
	}
	
	if (res == 0) {
		res = ahci_init_bus ();
	}
	if (res == 0) {
		res = ahci_init_root_hub (&ahci);
	}
	if (res != 0) {
		deinit_ahci ();
		return res;
	}
		
	PrintInfo ("AHCI ready\n");
	return 0;
} /* init_ahci */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void __exit exit_ahci(void) {	
	PrintInfo ("removing AHCI ...\n");
	
	deinit_ahci ();
	PrintInfo ("AHCI removed\n");
} /* exit_ahci */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
module_init(init_ahci);
module_exit(exit_ahci);

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

