/*---------------------------------------------------------------------------*\
 *   
 *   Copyright (C) 2006 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
\*---------------------------------------------------------------------------*/

#include <linux/interrupt.h>
#include "ahcidriver.h"
#include "ahcidef.h"
#include "ahcireg.h"

#include "ahciops.h"
#include "ahcilist.h"
#include "ahcirx.h"
#include "ahciport.h"
#include "ahcidesc.h"
#include "ahciurb.h"
#include "ahcicopy.h"
#include "ahcimem.h"

#include <asm/mach-ohio/hw_irq.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>

#include <asm/irq_cpu.h>
#include <asm/mipsregs.h>
#include <asm/system.h>
#include <asm/mips-boards/ohio.h>
#include <asm/mips-boards/ohioint.h>
#include <asm/mach-ohio/hw_irq.h>
#include <asm/mips-boards/ohio_vlynq.h>

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static AHCI_BYTE RxBuffer[CONFIG_AHCI_RX_MEMSIZE];

static volatile AHCI_DWORD g_oldRxAddr = 0;
/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

static unsigned ahci_Read_Fifo (char * RxBuffer);
static void ahci_ProcessRxBuffer (char * RxBuffer, unsigned RxBytes);
static void RefreshHPList (void);
static struct ahci_endpoint * ahci_irq_ep_get (struct usb_device* dev, unsigned int pipe);

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static int ahci_map_status (AHCI_TransferStatus ahci_status) {
	int usb_status;
	
	switch (ahci_status) {
	default:
		assert (0);
	case AHCI_TS_PidError:
	case AHCI_TS_OtherErrors:
	case AHCI_TS_EOF1Error:
		usb_status = USB_ST_INTERNALERROR;
		break;
		
	case AHCI_TS_NoError:
		usb_status = USB_ST_NOERROR;
		break;
	
	case AHCI_TS_CrcError:
		usb_status = USB_ST_CRC;
		break;
	
	case AHCI_TS_BitstuffError:
		usb_status = USB_ST_BITSTUFF;
		break;
	
	case AHCI_TS_Stall:
		usb_status = USB_ST_STALL;
		break;
	
	case AHCI_TS_Timeout:
		usb_status = USB_ST_NORESPONSE;
		break;
	
	case AHCI_TS_MissedEOP:
		usb_status = USB_ST_INTERNALERROR;
		break;
	
	case AHCI_TS_PacketOverflow:
		usb_status = USB_ST_DATAOVERRUN;
		break;
	
	case AHCI_TS_TransferOverflow:
		usb_status = USB_ST_BUFFEROVERRUN;
		break;
	
	case AHCI_TS_DataUnderrun:
		usb_status = USB_ST_SHORT_PACKET;
		break;
	
	case AHCI_TS_RxMemoryOverflow:
		//not really a status
		usb_status = USB_ST_URB_PENDING;
		break;
		
	case AHCI_TS_Canceled:
		usb_status = USB_ST_URB_KILLED;
		break;
	}
	
	return usb_status;
	
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static struct ahci_endpoint * ahci_irq_ep_get (struct usb_device* dev, unsigned int pipe) {
	struct ahci_endpoint * pahci_ep = NULL;

	struct usb_host_endpoint *ep = 
		(usb_pipein(pipe) ? dev->ep_in : dev->ep_out) [usb_pipeendpoint(pipe)];
	
	info (ep != NULL);

	if (ep != NULL) {
		spin_lock (&ahci.list_lock);
			pahci_ep = (struct ahci_ep *)ep->hcpriv;
		spin_unlock (&ahci.list_lock);
	}

	info (pahci_ep != NULL);
	return pahci_ep;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
/* irq handler */
irqreturn_t ahci_irq (int irq, void * dev_id, struct pt_regs *regs) {
	AHCI_DWORD	status;
	unsigned	bHostHalted	= 0;
	unsigned	count		= 0;
	irqreturn_t bHandled = IRQ_NONE;

	/* 
		Loop until IRQ status is stable, but at most 10 times
	*/
	while ((status = AHCI_READ_REGISTER (pAHCI_InterruptStatusReg)) & 
			(AHCI_ISR_hostHalted_FLAG|AHCI_ISR_portStatusChange_FLAG|AHCI_ISR_newStatusWord_FLAG)) {
		if (count++ > 10) {
			PrintInfo ("10 loops reached in AHCI IRQ! (%08X)\n", status);
			break;
		}

		/* W1C Flags */
		AHCI_WRITE_REGISTER (pAHCI_InterruptStatusReg, status);
		
		//PrintInfo ("IRQ for AHCI status %X\n", status);

		if (status & AHCI_ISR_hostHalted_FLAG) {
			/* Host Halted */
			PrintError ("Frame #%04X: Host Halted\n", AHCI_GetFrameNumber ());
			bHostHalted = 1;
			bHandled = IRQ_HANDLED;
		}
		
		if (status & AHCI_ISR_portStatusChange_FLAG) {
			PrintInfo ("Frame #%04X: PortStatusChanged\n", AHCI_GetFrameNumber ());
			ahci_UpdatePortStatus ();
			bHandled = IRQ_HANDLED;
		}

		if (status & AHCI_ISR_newStatusWord_FLAG) {
			unsigned RxBytes = ahci_Read_Fifo (RxBuffer);

			ahci_ProcessRxBuffer (RxBuffer, RxBytes);
			bHandled = IRQ_HANDLED;
		}
	}
	
	if (bHostHalted) {
		RefreshHPList ();
		PrintError ("Restart Host\n");
		AHCI_RestartHost ();
	}
	if (bHandled == IRQ_HANDLED) {
		tasklet_schedule (&ahci.completion_tasklet);
	}
	//PrintInfo ("IRQ for AHCI status danach %X\n", AHCI_READ_REGISTER (pAHCI_InterruptStatusReg));
	return bHandled;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
unsigned ahci_Read_Fifo (char * RxBuffer) {
	unsigned RxBytes = 0;

	unsigned curRxAddr = AHCI_GetCurRxAddress ();


	assert (RxBuffer != 0);
	
	//info (curRxAddr != g_oldRxAddr);
	
	if (curRxAddr != g_oldRxAddr) {
		//PrintInfo ("Pseudo DMATransfer curRxAddr == %X\n", g_curRxAddr);
		//info (g_curRxAddr > g_oldRxAddr);
		
		if (curRxAddr > g_oldRxAddr) {
			RxBytes = curRxAddr - g_oldRxAddr;

			assert ((RxBytes & 3) == 0);
			AHCI_DWORDCopyFromAlignedHW (RxBuffer, AHCI_PTR(g_oldRxAddr), RxBytes);
		} else {
			unsigned tmp_RxBytes = CONFIG_AHCI_RX_MEMSIZE - g_oldRxAddr;
			RxBytes = tmp_RxBytes + curRxAddr;

			assert ((RxBytes & 3) == 0);
			assert ((curRxAddr & 3) == 0);

			AHCI_DWORDCopyFromAlignedHW (RxBuffer, AHCI_PTR(g_oldRxAddr), tmp_RxBytes);
			AHCI_DWORDCopyFromAlignedHW (RxBuffer+tmp_RxBytes, AHCI_PTR(0), curRxAddr);
		}
		g_oldRxAddr = curRxAddr;
			//PrintInfo ("new curRxAddr == %X, Buffer: %u Bytes\n", curRxAddr, RxBytes);
	}
	return RxBytes;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void ahci_ProcessRxBuffer (char * RxBuffer, unsigned RxBytes) {
	assert ((RxBytes & 3) == 0);
	
	if (RxBytes) {
		unsigned pos = 0;

		while (RxBytes > pos) {
			unsigned StatusWord = *(unsigned *)(RxBuffer+pos);
			unsigned adr = (AHCI_TSW_tdPosition(StatusWord)<<4);

			assert (RxBytes >= pos + 4);
			assert (adr >= CONFIG_AHCI_RX_MEMSIZE);
			assert (adr < AHCI_TotalMemSize);
			
			if (adr < CONFIG_AHCI_RX_MEMSIZE) {
				PrintInfo ("\nerrRxBuffer[pos=%u] = %08X StatusWord for Descr@0x%X: completed=%u, status=%X Bytes = %u\n",
							pos, StatusWord,
							adr,
							AHCI_TSW_completed(StatusWord),
							AHCI_TSW_transferStatus(StatusWord),
							AHCI_TSW_byteCount(StatusWord)
							);


					PrintInfo ("Descriptor @0x%04X: %08X %08X %08X %08X\n",
						adr,
						AHCI_PTR(adr)[0],
						AHCI_PTR(adr)[1],
						AHCI_PTR(adr)[2],
						AHCI_PTR(adr)[3]);
			}

#			ifdef CONFIG_AHCI_DEBUG
			//if ((AHCI_TSW_transferStatus(StatusWord) != 0) ) 
			if ((AHCI_TSW_transferStatus(StatusWord) != 0) && 
					((AHCI_DEBUG_LEVEL > 1)||(AHCI_TSW_completed(StatusWord)))) {
					
				PrintInfo ("\nRxBuffer[pos=%u] = %08X StatusWord for Descr@0x%X: completed=%u, status=%X Bytes = %u urb %p\n",
							pos, StatusWord,
							adr,
							AHCI_TSW_completed(StatusWord),
							AHCI_TSW_transferStatus(StatusWord),
							AHCI_TSW_byteCount(StatusWord),
							AHCI_MemoryGetUrbPointer(adr));


				if (AHCI_TSW_completed(StatusWord)) {
					PrintInfo ("Descriptor @0x%04X: %08X %08X %08X %08X\n",
						adr,
						AHCI_PTR(adr)[0],
						AHCI_PTR(adr)[1],
						AHCI_PTR(adr)[2],
						AHCI_PTR(adr)[3]);
				}
			} 
#			endif

			pos += 4;
			{
				int status = ahci_map_status (AHCI_TSW_transferStatus(StatusWord));
				struct ahci_urb_ext * pUrbExt;
				urb_t * cur_urb = AHCI_MemoryGetUrbPointer(adr);

				assert (cur_urb != 0);
				pUrbExt = (struct ahci_urb_ext *)cur_urb->hcpriv;

				assert (pUrbExt != 0);
				assert (pUrbExt->td_done < pUrbExt->td_count);

				assert (adr == pUrbExt->td_adr[pUrbExt->td_done]);

				assert (RxBytes >= (pos + AHCI_TSW_byteCount(StatusWord)));
					
				if (AHCI_TSW_transferStatus(StatusWord) != AHCI_TS_RxMemoryOverflow) {
					unsigned nBytes = AHCI_TSW_byteCount(StatusWord);
					if (nBytes > 0) {
						unsigned len = nBytes;

						assert (usb_pipein (cur_urb->pipe));
						if (usb_pipeisoc (cur_urb->pipe)) {
								assert (pUrbExt->iso_done < cur_urb->number_of_packets);

								assert (nBytes <= cur_urb->iso_frame_desc[pUrbExt->iso_done].length);

								memcpy (((char*)cur_urb->transfer_buffer)+cur_urb->iso_frame_desc[pUrbExt->iso_done].offset,
										RxBuffer+pos, nBytes);

								cur_urb->iso_frame_desc[pUrbExt->iso_done].actual_length = nBytes;

						} else {
							if (AHCI_TSW_transferStatus(StatusWord) == AHCI_TS_TransferOverflow) {
								len = cur_urb->transfer_buffer_length - cur_urb->actual_length;
								assert (len < nBytes);
							}
							assert (cur_urb->actual_length + len <= cur_urb->transfer_buffer_length);
							memcpy (((char*)cur_urb->transfer_buffer)+cur_urb->actual_length,
										RxBuffer+pos, len);
						}
						cur_urb->actual_length += len;

						pos += ((nBytes +3) & (~3));

					}
					assert (pos <= RxBytes);

					if (usb_pipeisoc (cur_urb->pipe)) {
						if (pUrbExt->td_count > 1 || usb_pipein (cur_urb->pipe)) {
							cur_urb->iso_frame_desc[pUrbExt->iso_done].status = status;
							pUrbExt->iso_done ++;
							
							if (status != USB_ST_NOERROR) {
								cur_urb->error_count ++;
							}
						} else {
							unsigned i;
							unsigned sublen = cur_urb->iso_frame_desc[0].length;
							assert (nBytes == 0);
							assert (usb_pipeout (cur_urb->pipe));
							assert (status == USB_ST_NOERROR);

							for (i = 0; i < cur_urb->number_of_packets; i++) {
								assert (sublen == cur_urb->iso_frame_desc[i].length);
								cur_urb->iso_frame_desc[i].actual_length = sublen;
								cur_urb->iso_frame_desc[i].status = status;
								cur_urb->actual_length += sublen;
							}
						}
					}
				}

				if (AHCI_TSW_completed(StatusWord)) {
					unsigned ed = usb_pipeendpoint (cur_urb->pipe);
					unsigned last_adr = adr;
					unsigned free_count = 1;
					
					assert (status != USB_ST_URB_PENDING);
					assert (AHCI_TSW_transferStatus(StatusWord) != AHCI_TS_RxMemoryOverflow);
					
					if (AHCI_TSW_transferStatus(StatusWord) && (AHCI_TSW_transferStatus(StatusWord) != 0xF)) {
						if (pUrbExt->tds_allocated == pUrbExt->td_count) {
							unsigned nV = AHCI_DWORD4_nextVertical(AHCI_PTR(pUrbExt->td_adr[pUrbExt->td_count-1])[3]);
							//Existiert Nachfolger?
							if (nV != 0) {
								//to do :: toggle ???

								nV = nV <<4;
								PrintInfo ("Reinsert nextV %x\n", nV);
								info (AHCI_DWORD1_period(AHCI_PTR(adr)[0]) == 0);
								AHCI_InsertDescriptor (AHCI_DWORD1_period(AHCI_PTR(adr)[0])?
												AHCI_LIST_highPrio:
												AHCI_LIST_lowPrio, nV);
							}
						}

						pUrbExt->td_adr[pUrbExt->td_done] = 0;
						pUrbExt->td_done ++;
						
						for (; pUrbExt->td_done < pUrbExt->tds_allocated; pUrbExt->td_done++) {
							last_adr = pUrbExt->td_adr[pUrbExt->td_done];
							assert (last_adr != 0);
							
							PrintInfo ("urb %p:Free following Desc %x (index %u) after Error\n", cur_urb, last_adr, pUrbExt->td_done);
							assert (AHCI_MemoryGetUrbPointer(last_adr) == cur_urb);
							assert (!AHCI_DWORD4_nextHorizontal(AHCI_PTR(last_adr)[3]));
							assert (!AHCI_DWORD4_completed(AHCI_PTR(last_adr)[3]));
							
							pUrbExt->td_adr[pUrbExt->td_done] = 0;
							AHCI_MemoryFree (last_adr);
							free_count ++;
						}

					} else {
						pUrbExt->td_adr[pUrbExt->td_done] = 0;
						pUrbExt->td_done ++;
					}
					
					//actual_length fr OUT berechnen
					if (usb_pipeout (cur_urb->pipe) && !usb_pipeisoc (cur_urb->pipe) 
									&& (cur_urb->transfer_buffer_length != 0)) {
						if (!usb_pipecontrol (cur_urb->pipe) || 
							((pUrbExt->td_done > 0) && pUrbExt->td_done < pUrbExt->td_count)) {
								
							unsigned addlen = AHCI_MAX_OUT_BLOCK_SIZE;
							unsigned td_max = usb_pipecontrol (cur_urb->pipe) 
									? pUrbExt->td_count-1
									: pUrbExt->td_count;
									
							if (pUrbExt->td_done == td_max) {
								addlen = (cur_urb->transfer_buffer_length%AHCI_MAX_OUT_BLOCK_SIZE);
								if (addlen == 0) {
									addlen = AHCI_MAX_OUT_BLOCK_SIZE;
								}
							}
							if (AHCI_TSW_transferStatus(StatusWord)) {
								unsigned restlen = AHCI_DWORD3_transferSize (AHCI_PTR(adr)[2]);
								addlen -= restlen;
							}

							cur_urb->actual_length += addlen;
						}
						
					
					}

					{
						struct ahci_endpoint * pahci_ep = ahci_irq_ep_get (cur_urb->dev, cur_urb->pipe);
						info (pahci_ep != NULL);
												
						if (pahci_ep != NULL) {
							if (pUrbExt->td_done == pUrbExt->tds_allocated) {
								if (pahci_ep->last_adr == last_adr) {
									pahci_ep->last_adr = 0;
								}
								if (!usb_pipeisoc (cur_urb->pipe)) {
									unsigned toggle = AHCI_TSW_toggleBit(StatusWord);
									usb_settoggle (cur_urb->dev, ed, usb_pipeout (cur_urb->pipe), toggle);
								}
							}
							assert (pahci_ep->pending_tds >= free_count);
							pahci_ep->pending_tds-= free_count;
						}
					}
					AHCI_MemoryFree (adr);
					
					assert ((pUrbExt->td_done < pUrbExt->td_count) || (pUrbExt->td_done == pUrbExt->tds_allocated));
					if (pUrbExt->td_done == pUrbExt->tds_allocated) {
						if (status == USB_ST_NOERROR) {
							if (pUrbExt->td_done < pUrbExt->td_count) {
								if (pUrbExt->unlink_called) {
									status = USB_ST_URB_KILLED;
								} else {
									status = USB_ST_URB_PENDING;
								}
							}
						}
						
						assert ((status != USB_ST_URB_KILLED) || (pUrbExt->unlink_called));

						if (status != USB_ST_URB_PENDING) {
							pUrbExt->status = status;

							 //PrintInfo ("cur_urb %p complete with actual_length %u status %d\n",
							//		cur_urb, cur_urb->actual_length, cur_urb->status);

							spin_lock (&ahci.list_lock);
							if (pUrbExt->td_done == pUrbExt->td_count) {
								list_add_tail (&cur_urb->urb_list, &ahci.completion_list);
							}
							spin_unlock (&ahci.list_lock);
						}
					}
					
				} /*completed*/
			}/*block*/
		}/*while*/
		
	} /*rxbytes*/ 
}


/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void RefreshHPList (void) {
	unsigned adr = AHCI_READ_REGISTER (pAHCI_HPStartPointerReg);
	unsigned CurrentFrame = AHCI_GetFrameNumber () + 3; 
	
#	ifdef CONFIG_AHCI_DEBUG
	unsigned loopcount = 0;
#	endif

	while (adr) {
		unsigned TimedFrame;

		//TestPrint ("update adr %X with TimedFrame %X\n", adr, TimedFrame);
#		ifdef CONFIG_AHCI_DEBUG
		loopcount ++;

		if (loopcount > 100) {
			assert (0);
		 	break;
		}
#		endif

		assert (adr >= CONFIG_AHCI_RX_MEMSIZE);
		assert (adr < AHCI_TotalMemSize);
		
		TimedFrame = AHCI_PTR(adr)[1] & 0xFFFF;
		
		if ((TimedFrame - CurrentFrame) > 10000) {
			if (AHCI_DWORD2_isochron (AHCI_PTR(adr)[1])) {
				unsigned nVAdr = adr;
				PrintInfo ("Abort ISO\n");
				do {
					AHCI_RemoveDescriptor (nVAdr);
					nVAdr = (AHCI_DWORD4_nextVertical(AHCI_PTR(nVAdr)[3]) << 4);
				} while (nVAdr != 0);
				
			} else {
				unsigned period = AHCI_DWORD1_period (AHCI_PTR(adr)[0]);

#				if defined CONFIG_AHCI_DEBUG			
				unsigned iloopcount = 0 ;
#				endif			
				
				while ((TimedFrame - CurrentFrame) > 10000) {
#					if defined CONFIG_AHCI_DEBUG			
					iloopcount ++;

					if (iloopcount > 100) {
						assert (0);
						break;
					}
#					endif			

					TimedFrame = (TimedFrame + period) & 0xFFFF;
				}
				PrintInfo ("updated adr %X to Frame %X\n", adr, TimedFrame);

				assert ((TimedFrame - CurrentFrame) < 9000);
				AHCI_PTR(adr)[1] = (AHCI_PTR(adr)[1] & 0xFFFF0000) | (TimedFrame);
			}
		}
		
		adr = AHCI_DWORD4_nextHorizontal(AHCI_PTR(adr)[3])<<4;
	}
}

