/*---------------------------------------------------------------------------*\
 *   
 *   Copyright (C) 2006 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
\*---------------------------------------------------------------------------*/

#include "ahcidriver.h"
#include "ahcidef.h"
#include "ahcireg.h"


#include "ahcimem.h"

#include <linux/list.h>

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
#define AHCI_BLOCK_SIZE_POWER (4)
#define AHCI_BLOCK_SIZE (1<<AHCI_BLOCK_SIZE_POWER)

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
typedef struct _ahci_mem_block {
	unsigned urb_pointer;
	unsigned num_blocks;
	unsigned is_free;
	unsigned next_free;
} ahci_mem_block;

static spinlock_t s_ahci_MemLock;
static ahci_mem_block *s_AHCI_MemBlocks;
static int s_iFreeStart;
static unsigned s_uStartAddress;
static unsigned s_nMaxBlocks = 0;

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void AHCI_MemorySetUrbPointer (unsigned Adr, urb_t *pUrb) {
	int index = (Adr - s_uStartAddress) >> AHCI_BLOCK_SIZE_POWER;

	assert (s_AHCI_MemBlocks[index].is_free == 0);
	assert (s_AHCI_MemBlocks[index].urb_pointer == 0);
	s_AHCI_MemBlocks[index].urb_pointer = (unsigned) pUrb;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
urb_t * AHCI_MemoryGetUrbPointer (unsigned Adr) {
	int index = (Adr - s_uStartAddress) >> AHCI_BLOCK_SIZE_POWER;

	assert (s_AHCI_MemBlocks[index].is_free == 0);
	info (s_AHCI_MemBlocks[index].urb_pointer != 0);
	
	return (urb_t *) s_AHCI_MemBlocks[index].urb_pointer;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
int AHCI_MemoryInit (unsigned uStartAddress, size_t size) {
	assert (size >= 4096);

	s_uStartAddress = uStartAddress;
	s_iFreeStart = 0;
	s_nMaxBlocks = size/AHCI_BLOCK_SIZE;
	
	assert (AHCI_BLOCK_SIZE >= 16);
	assert (AHCI_BLOCK_SIZE <= 256);

	spin_lock_init(&s_ahci_MemLock);

	info (uStartAddress >= CONFIG_AHCI_RX_MEMSIZE);
	if (uStartAddress < CONFIG_AHCI_RX_MEMSIZE) {
		return -1;
	}
	
	info ((uStartAddress % 16) == 0);
	if ((uStartAddress % 16) != 0) {
		return -1;
	}

	info (size >= 4096);
	if (size < 4096) {
		return -1;
	}

	s_AHCI_MemBlocks = kmalloc (s_nMaxBlocks * sizeof(ahci_mem_block), GFP_KERNEL);
	
	info (s_AHCI_MemBlocks != 0);
	
	if (s_AHCI_MemBlocks == 0) {
		return -1;
	}

	s_AHCI_MemBlocks[0].num_blocks = s_nMaxBlocks;
	s_AHCI_MemBlocks[0].urb_pointer = 0;
	s_AHCI_MemBlocks[0].is_free = 1;
	s_AHCI_MemBlocks[0].next_free = -1;

	PrintInfo ("AHCIMem: %u Blocks of size %u, struct size is %u\n", s_nMaxBlocks, AHCI_BLOCK_SIZE, sizeof(ahci_mem_block));

	return 0;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void AHCI_MemoryDeinit (void) {
	if (s_AHCI_MemBlocks != 0) {
		assert (s_AHCI_MemBlocks[0].num_blocks == s_nMaxBlocks);
		assert (s_AHCI_MemBlocks[0].is_free);
		kfree (s_AHCI_MemBlocks);
	}
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
unsigned AHCI_MemoryAlloc (size_t size) {
	unsigned blocks = (size +(AHCI_BLOCK_SIZE-1)) >> AHCI_BLOCK_SIZE_POWER;
	unsigned rest_blocks = 0;
	int i = -1;
	int prev_free = -1;
	
	unsigned long flags;
	
#	ifdef	CONFIG_AHCI_DEBUG
	unsigned max_block_size = 0;
	unsigned block_count = 0;

	unsigned count = 0;
#	endif
	
	info (blocks > 0);
	if (blocks == 0) {
		blocks = 1;
	}
	//PrintInfo ("AHCI_ALLOC Enter\n");
	spin_lock_irqsave (&s_ahci_MemLock, flags);
	i = s_iFreeStart;
	while (i >= 0) {
		assert (s_AHCI_MemBlocks[i].is_free);
		assert (s_AHCI_MemBlocks[i].num_blocks != 0);
		if (s_AHCI_MemBlocks[i].num_blocks >= blocks) {
			rest_blocks = s_AHCI_MemBlocks[i].num_blocks - blocks;
			break;
		}

#		ifdef	CONFIG_AHCI_DEBUG
		count ++;
		assert (count < 1000);
		if (count >= 1000) {
			i = -1;
			break;
		}
#		endif
		
#		ifdef	CONFIG_AHCI_DEBUG
		block_count ++;
		if (max_block_size < s_AHCI_MemBlocks[i].num_blocks) {
			max_block_size = s_AHCI_MemBlocks[i].num_blocks;
		}
#		endif

		prev_free = i;
		i = s_AHCI_MemBlocks[i].next_free;
	}
	if (i < 0) {
		spin_unlock_irqrestore (&s_ahci_MemLock, flags);
		PrintError ("no AHCI_DESC MEM for size %u, %u blocks [max block is %u]\n", 
					size, block_count, max_block_size<<AHCI_BLOCK_SIZE_POWER);
		return 0;
	}

	if (rest_blocks > 0) {
		s_AHCI_MemBlocks[i+blocks].num_blocks = rest_blocks;
		s_AHCI_MemBlocks[i+blocks].is_free = 1;
		s_AHCI_MemBlocks[i+blocks].next_free = s_AHCI_MemBlocks[i].next_free;
		
		if (prev_free == -1) {
			assert (s_iFreeStart == i);
			s_iFreeStart = i + blocks;
		} else {
			s_AHCI_MemBlocks[prev_free].next_free = i + blocks;
		}
	} else {
		if (prev_free == -1) {
			assert (s_iFreeStart == i);
			s_iFreeStart = s_AHCI_MemBlocks[i].next_free;
		} else {
			s_AHCI_MemBlocks[prev_free].next_free = s_AHCI_MemBlocks[i].next_free;
		}
	}
	
	s_AHCI_MemBlocks[i].next_free = 0;
	s_AHCI_MemBlocks[i].is_free = 0;
	s_AHCI_MemBlocks[i].num_blocks = blocks;
	s_AHCI_MemBlocks[i].urb_pointer = 0;
	
	spin_unlock_irqrestore (&s_ahci_MemLock, flags);
	
	//PrintInfo ("alloc index i = %u blocks = %u\n", i, blocks);
	//PrintInfo ("AHCI_ALLOC Leave %X\n", (i<<AHCI_BLOCK_SIZE_POWER) + (s_uStartAddress));
	return (i<<AHCI_BLOCK_SIZE_POWER) + (s_uStartAddress);
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void AHCI_MemoryFree (unsigned Adr) {
	unsigned long flags;
	int index = (Adr - s_uStartAddress) >> AHCI_BLOCK_SIZE_POWER;

	int next_free = -1;
	//PrintInfo ("free index i = %u blocks = %u\n", index, s_AHCI_MemBlocks[index].num_blocks);
	
	assert (s_AHCI_MemBlocks[index].is_free == 0);

	spin_lock_irqsave (&s_ahci_MemLock, flags);
	next_free = s_iFreeStart;
	
	if (index < s_iFreeStart) {
		info (s_iFreeStart != -1);
		s_iFreeStart = index;
		s_AHCI_MemBlocks[index].is_free = 1;
		s_AHCI_MemBlocks[index].next_free = next_free;
	} else {
		int prev = 0xCAFEBABE;
		
#		ifdef	CONFIG_AHCI_DEBUG
		unsigned count = 0;
#		endif		
		
		assert (next_free != -1);
		assert (index > s_iFreeStart);
		
		do {
#			ifdef	CONFIG_AHCI_DEBUG
			count ++;
			assert (count < 1000);
			if (count >= 1000) {
				PrintInfo ("Free count too Big\n");
				PrintInfo ("Free_Start = %X\n", s_iFreeStart);
				PrintInfo ("index = %X, prev = %X next = %X\n", index, prev, next_free);
				break;
			}
#			endif

			prev = next_free;
			assert (s_AHCI_MemBlocks[prev].is_free);

			next_free = s_AHCI_MemBlocks[prev].next_free;
			
			assert ((next_free > prev) || (next_free == -1));
			if (next_free > index) {
				break;
			}

		} while (next_free != -1);
		
		//PrintInfo ("Free middle prev = %d next = %d\n",prev, next_free);
		if (s_AHCI_MemBlocks[prev].num_blocks + prev == index) {
			s_AHCI_MemBlocks[prev].num_blocks += s_AHCI_MemBlocks[index].num_blocks;
			index = prev;
		} else {
			s_AHCI_MemBlocks[prev].next_free = index;
			s_AHCI_MemBlocks[index].next_free = next_free;
			s_AHCI_MemBlocks[index].is_free = 1;
		}
	}
	
	assert (s_AHCI_MemBlocks[index].next_free == next_free);
	
	if (next_free == index + s_AHCI_MemBlocks[index].num_blocks) {
		assert (next_free != -1);
		
		s_AHCI_MemBlocks[index].num_blocks += s_AHCI_MemBlocks[next_free].num_blocks;
		s_AHCI_MemBlocks[index].next_free = s_AHCI_MemBlocks[next_free].next_free;


		s_AHCI_MemBlocks[next_free].num_blocks = 0;
		s_AHCI_MemBlocks[next_free].is_free = 0;
		s_AHCI_MemBlocks[next_free].next_free = -1;
	}
	spin_unlock_irqrestore (&s_ahci_MemLock, flags);
	
	//PrintInfo ("Free done s_iFreeStart:%u \n",s_iFreeStart);
	//ahci_DebugPrintf("Free done s_iFreeStart:%u \n",s_iFreeStart);
	assert (s_AHCI_MemBlocks[index].is_free);
	
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
