/*---------------------------------------------------------------------------*\
 *   
 *   Copyright (C) 2006 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
\*---------------------------------------------------------------------------*/

#include <linux/list.h>
#include "ahcidef.h"
#include "ahcidriver.h"
#include "ahciroothub.h"
#include "ahciurb.h"
#include "ahcidesc.h"
#include "ahciops.h"
#include "ahcilist.h"
#include "ahcimem.h"
#include "ahcicopy.h"

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

/* Forward declarations to silence the compiler... */

static int ahci_submit_control_urb (urb_t * urb);
static int ahci_submit_iso_urb (urb_t * urb);
static int ahci_submit_int_urb (urb_t * urb);
static int ahci_submit_bulk_urb (urb_t * urb);


static void process_control_urb (urb_t * urb);
//static void process_iso_urb (urb_t * urb);
static void process_int_urb (urb_t * urb);
static void process_bulk_urb (urb_t * urb);

static void process_control_list (void);
//static void process_iso_list (void);
static void process_int_list (void);
static void process_bulk_list (void);

extern struct pt_regs	dummy_regs;	/* See: ahciroothub.c */

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static void unlink_special (struct ahci_urb_ext * pUrbExt) {
	info (!pUrbExt->unlink_called);
	if (!pUrbExt->unlink_called) {
		unsigned count;
		unsigned adr;
		unsigned flags;

		spin_lock_irqsave (&ahci.list_lock, flags);
		pUrbExt->unlink_called = 1;
		count = pUrbExt->tds_allocated;

		PrintInfo ("ahci_unlink_special: td_count = %u tds_allocated = %u td_done = %u called\n",pUrbExt->td_count, pUrbExt->tds_allocated, pUrbExt->td_done);
		while (pUrbExt->td_done < count) {
			count --;
			adr = pUrbExt->td_adr[count];
			assert (adr != 0);
			//assert (AHCI_MemoryGetUrbPointer(adr) == urb);
			PrintInfo ("AHCI_RemoveDescriptor @0x%04X: %08X %08X %08X %08X\n",
						adr,
						AHCI_PTR(adr)[0],
						AHCI_PTR(adr)[1],
						AHCI_PTR(adr)[2],
						AHCI_PTR(adr)[3]);

			AHCI_RemoveDescriptor (adr);
		}
		spin_unlock_irqrestore (&ahci.list_lock, flags);

		tasklet_schedule (&ahci.completion_tasklet);
	}
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static struct ahci_endpoint * ahci_ep_get (struct usb_device* dev, unsigned int pipe) {
	struct ahci_endpoint * pahci_ep = NULL;
	struct usb_host_endpoint *ep = 
		(usb_pipein(pipe) ? dev->ep_in : dev->ep_out) [usb_pipeendpoint(pipe)];
		
	info (ep != NULL);
	if (ep != NULL) {
		unsigned flags;
		spin_lock_irqsave (&ahci.list_lock, flags);
		pahci_ep = (struct ahci_ep *)ep->hcpriv;
		if (pahci_ep == NULL) {
			pahci_ep = kmalloc (sizeof (struct ahci_endpoint), GFP_ATOMIC);
			PrintInfo ("new ahci_endpoint %p on dev#%u\n", pahci_ep, dev->devnum);
			if (pahci_ep != NULL) {
				memset (pahci_ep, 0, sizeof (struct ahci_endpoint));
				ep->hcpriv = pahci_ep;
			}
		}
		spin_unlock_irqrestore (&ahci.list_lock, flags);
	}
	info (pahci_ep != NULL);
	return pahci_ep;
}


/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
void ahci_endpoint_disable (struct usb_device *dev, struct usb_host_endpoint *ep) {
	struct ahci_endpoint * pahci_ep = (struct ahci_endpoint *)ep->hcpriv;
	PrintInfo ("ahci_endpoint_disable %p on dev#%u\n", pahci_ep, dev->devnum);
	if (pahci_ep != NULL) {
		kfree (pahci_ep);
	}
	ep->hcpriv = NULL;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static struct ahci_urb_ext *ahci_alloc_urb_ext (unsigned nDescriptors) {
	unsigned length = sizeof(struct ahci_urb_ext) + nDescriptors * sizeof(AHCI_DESC_ADR);
	struct ahci_urb_ext *pUrbExt = kmalloc (length, GFP_ATOMIC);
	
	if (pUrbExt) {
		memset (pUrbExt, 0, length);
		pUrbExt->td_count = nDescriptors;
	}
	return pUrbExt;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static void ahci_free_urb_ext (urb_t *urb) {
	unsigned i;
	struct ahci_urb_ext *pUrbExt = urb->hcpriv;
	urb->hcpriv = 0;
	
	info (pUrbExt != 0);
	if (pUrbExt != 0) {
		for (i = 0; i < pUrbExt->td_count; i ++) {
			unsigned adr = pUrbExt->td_adr[i];

			info (adr == 0 || i >= pUrbExt->tds_allocated);
			if (adr != 0) {
				AHCI_MemoryFree (adr);
			}
		}
		kfree (pUrbExt);
	}
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
int ahci_submit_urb (urb_t * urb, unsigned mem_flags) {
	/* PrintInfo ("ahci_submit_urb %p, %08x called\n", urb, mem_flags); */
	
	if (!urb->dev || !urb->dev->bus)
		return -ENODEV;
	
	info (urb->hcpriv == 0);
	if (urb->hcpriv)			/* urb already in use */
		return -EINVAL;

	/* handle a request to the virtual root hub */
	if (usb_pipedevice (urb->pipe) == ahci.rh_address) {
		return ahci_rh_submit_urb (urb);
	}
	
	/* for the private part of the URB we need the number of TDs (size) */
	switch (usb_pipetype (urb->pipe)) {
		case PIPE_BULK:
			/* PrintInfo ("ahci submit BULK urb %p\n", urb); */
			return ahci_submit_bulk_urb (urb);
			break;
		case PIPE_ISOCHRONOUS:
			/* PrintInfo ("ahci submit ISO urb %p\n", urb); */
			return ahci_submit_iso_urb (urb);
			break;
		case PIPE_CONTROL:
			/* PrintInfo ("ahci submit CONTROL urb %p\n", urb); */
			return ahci_submit_control_urb (urb);
			break;
		case PIPE_INTERRUPT:
			/* PrintInfo ("ahci submit INT urb %p\n", urb); */
			return ahci_submit_int_urb (urb);
			break;
	}
	
	return -EINVAL;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
int ahci_unlink_urb (urb_t * urb, int status) {
	struct ahci_urb_ext * pUrbExt;
	
	/* PrintInfo ("ahci_unlink_urb %p, %d called\n", urb, status); */
	if (!urb) /* just to be sure */ 
		return -EINVAL;
		
	if (!urb->dev || !urb->dev->bus)
		return -ENODEV;

	if (!(urb->transfer_flags & URB_ASYNC_UNLINK) && in_interrupt ()) {
		return -EWOULDBLOCK;
	}

	/* handle a request to the virtual root hub */
	if (usb_pipedevice (urb->pipe) == ahci.rh_address) {
		return ahci_rh_unlink_urb (urb);
	}

	pUrbExt = urb->hcpriv;
	info (pUrbExt != 0);
	if (pUrbExt == 0) {
			return -EINVAL;
	}
	
	info (!pUrbExt->unlink_called);
	if (!pUrbExt->unlink_called) {
		unsigned count;
		unsigned adr;
		unsigned flags;

		spin_lock_irqsave (&ahci.list_lock, flags);
		pUrbExt->unlink_called = 1;
		count = pUrbExt->tds_allocated;

		PrintInfo ("ahci_unlink_urb: td_count = %u tds_allocated = %u td_done = %u called\n",pUrbExt->td_count, pUrbExt->tds_allocated, pUrbExt->td_done);
		while (pUrbExt->td_done < count) {
			count --;
			adr = pUrbExt->td_adr[count];
			assert (adr != 0);
			assert (AHCI_MemoryGetUrbPointer(adr) == urb);
			PrintInfo ("AHCI_RemoveDescriptor @0x%04X: %08X %08X %08X %08X\n",
						adr,
						AHCI_PTR(adr)[0],
						AHCI_PTR(adr)[1],
						AHCI_PTR(adr)[2],
						AHCI_PTR(adr)[3]);
			
			AHCI_RemoveDescriptor (adr);
		}
		spin_unlock_irqrestore (&ahci.list_lock, flags);

		tasklet_schedule (&ahci.completion_tasklet);

		
		if (!(urb->transfer_flags & URB_ASYNC_UNLINK)) {
			int timeout = 200;
			DECLARE_WAIT_QUEUE_HEAD (wait_head); 
			DECLARE_WAITQUEUE (wait_elem, current);

			add_wait_queue (&wait_head, &wait_elem);
			pUrbExt->wait_head = &wait_head;
			PrintInfo ("ahci_unlink_urb %p wait\n", urb);

			/* wait until all TDs are deleted */
			set_current_state(TASK_UNINTERRUPTIBLE);

			while (timeout && (urb->status == USB_ST_URB_PENDING)) {
				timeout = schedule_timeout (timeout);
			}
			set_current_state(TASK_RUNNING);
			remove_wait_queue (&wait_head, &wait_elem);
			PrintInfo ("ahci_unlink_urb %p st = %d wait done\n", urb, urb->status);
			if (urb->status == USB_ST_URB_PENDING) {
				return -ETIMEDOUT;
			}
		}
	}
	
	PrintInfo ("ahci_unlink_urb %p done\n", urb);
	
	return 0;

}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static int ahci_submit_control_urb (urb_t * urb) {
	unsigned pipe = urb->pipe;
	unsigned len = urb->transfer_buffer_length;

	struct ahci_urb_ext * pUrbExt;

	unsigned max_tds = 2;

	info (len <= (8*1024));
	//PrintInfo ("control: urb %p len = %u out = %u\n", urb, len, usb_pipeout (pipe));
	assert (usb_pipecontrol(pipe));

	if (len > 0) {
		max_tds ++;
		if (usb_pipein(pipe))  {
			if (len > AHCI_MAX_IN_BLOCK_SIZE) {
				max_tds += (len-1)/AHCI_MAX_IN_BLOCK_SIZE;
			}
		} else {
			if (len > AHCI_MAX_OUT_BLOCK_SIZE) {
				max_tds += (len-1)/AHCI_MAX_OUT_BLOCK_SIZE;
			}
		}
	}

	pUrbExt = ahci_alloc_urb_ext (max_tds);
	info (pUrbExt != 0);
	if (pUrbExt == 0) {
		return -ENOMEM;
	}
	urb->status = USB_ST_URB_PENDING;
	urb->actual_length = 0;
	urb->hcpriv = pUrbExt;

    {
        unsigned flags;
        spin_lock_irqsave (&ahci.list_lock, flags);
            list_add_tail (&urb->urb_list, &ahci.control_pending_list);
        spin_unlock_irqrestore (&ahci.list_lock, flags);    
    }

	tasklet_schedule (&ahci.completion_tasklet);
	
	return 0;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static int ahci_submit_iso_urb (urb_t * urb) {
	return -EINVAL;
/*
	FIXME
	
	unsigned pipe = urb->pipe;

	struct ahci_urb_ext * pUrbExt;

	unsigned max_tds = 1;

	//PrintInfo ("iso: maxpacket = %u len = %u out = %u toggle %u\n", maxpacket, len, usb_pipeout (pipe), toggle);
	assert (usb_pipeisoc(pipe));

	if (usb_pipein(pipe))  {
		unsigned pipe = urb->pipe;
		unsigned maxpacket = usb_maxpacket (urb->dev, pipe, usb_pipeout (pipe));
		unsigned len = maxpacket*urb->number_of_packets;
		if (len > AHCI_MAX_IN_BLOCK_SIZE) {
			max_tds += (len-1)/AHCI_MAX_IN_BLOCK_SIZE;
		}
	} else {
		max_tds = urb->number_of_packets;
	}

	pUrbExt = ahci_alloc_urb_ext (max_tds);
	info (pUrbExt != 0);
	if (pUrbExt == 0) {
		return -ENOMEM;
	}
	urb->status = USB_ST_URB_PENDING;
	urb->actual_length = 0;
	urb->hcpriv = pUrbExt;

    {
        unsigned flags;
        spin_lock_irqsave (&ahci.list_lock, flags);
            list_add_tail (&urb->urb_list, &ahci.iso_pending_list);
        spin_unlock_irqrestore (&ahci.list_lock, flags);    
    }

	tasklet_schedule (&ahci.completion_tasklet);
	
	return 0;
*/
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static int ahci_submit_int_urb (urb_t * urb) {
#	if defined (CONFIG_AHCI_DEBUG)
	unsigned pipe = urb->pipe;
#	endif	
	unsigned len = urb->transfer_buffer_length;

	struct ahci_urb_ext * pUrbExt;

	unsigned max_tds = 1;

	info (len <= (8*1024));
	//PrintInfo ("bulk: maxpacket = %u len = %u out = %u toggle %u\n", maxpacket, len, usb_pipeout (pipe), toggle);
	assert (usb_pipeint(pipe));

	assert (usb_pipein(pipe));
	
	if (len > AHCI_MAX_IN_BLOCK_SIZE) {
		max_tds += (len-1)/AHCI_MAX_IN_BLOCK_SIZE;
	}

	pUrbExt = ahci_alloc_urb_ext (max_tds);

	info (pUrbExt != 0);
	if (pUrbExt == 0) {
		return -ENOMEM;
	}

	urb->status = USB_ST_URB_PENDING;
	urb->actual_length = 0;
	urb->hcpriv = pUrbExt;

    {
        unsigned flags;
        spin_lock_irqsave (&ahci.list_lock, flags);
            list_add_tail (&urb->urb_list, &ahci.int_pending_list);
        spin_unlock_irqrestore (&ahci.list_lock, flags);    
    }

	tasklet_schedule (&ahci.completion_tasklet);
	
	return 0;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static int ahci_submit_bulk_urb (urb_t * urb) {
	unsigned pipe = urb->pipe;
	unsigned len = urb->transfer_buffer_length;

	struct ahci_urb_ext * pUrbExt;

	unsigned max_tds = 1;

	//info (len <= (8*1024));
	
	//if (len > (8*1024)) {
	//	PrintInfo ("bulk: len = %u out = %u\n", len, usb_pipeout (pipe));
	//}
	assert (usb_pipebulk(pipe));

	if (usb_pipein(pipe))  {
		if (len > AHCI_MAX_IN_BLOCK_SIZE) {
			max_tds += (len-1)/AHCI_MAX_IN_BLOCK_SIZE;
		}
	} else {
		if (len > AHCI_MAX_OUT_BLOCK_SIZE) {
			max_tds += (len-1)/AHCI_MAX_OUT_BLOCK_SIZE;
		}
	}

	pUrbExt = ahci_alloc_urb_ext (max_tds);
	info (pUrbExt != 0);
	if (pUrbExt == 0) {
		return -ENOMEM;
	}
	urb->status = USB_ST_URB_PENDING;
	urb->actual_length = 0;
	urb->hcpriv = pUrbExt;

    {
        unsigned flags;
        spin_lock_irqsave (&ahci.list_lock, flags);
            list_add_tail (&urb->urb_list, &ahci.bulk_pending_list);
        spin_unlock_irqrestore (&ahci.list_lock, flags);    
    }

	tasklet_schedule (&ahci.completion_tasklet);
	
	return 0;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static void process_control_list (void) {
	unsigned flags;
	
	struct list_head *pos;
	struct list_head *temp;
	urb_t * urb;
	
	spin_lock_irqsave (&ahci.list_lock, flags);
	
	list_for_each_safe (pos, temp, &ahci.control_pending_list) {

		urb = list_entry (pos, urb_t, urb_list);
		
		assert (&urb->urb_list == pos->prev->next);
		assert (urb->hcpriv != 0);

        spin_unlock_irqrestore (&ahci.list_lock, flags);
		
		process_control_urb (urb);
		
		spin_lock_irqsave (&ahci.list_lock, flags);
	}
	
	spin_unlock_irqrestore (&ahci.list_lock, flags);	
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static void process_bulk_list (void) {
	unsigned flags;
	
	struct list_head *pos;
	struct list_head *temp;
	urb_t * urb;

	spin_lock_irqsave (&ahci.list_lock, flags);
	
	list_for_each_safe (pos, temp, &ahci.bulk_pending_list) {

		urb = list_entry (pos, urb_t, urb_list);
		
		assert (&urb->urb_list == pos->prev->next);
		assert (urb->hcpriv != 0);

        spin_unlock_irqrestore (&ahci.list_lock, flags);
		
		process_bulk_urb (urb);

		spin_lock_irqsave (&ahci.list_lock, flags);
	}
	
	spin_unlock_irqrestore (&ahci.list_lock, flags);	
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static void process_int_list (void) {
	unsigned flags;
	
	struct list_head *pos;
	struct list_head *temp;
	urb_t * urb;

	spin_lock_irqsave (&ahci.list_lock, flags);
	
	list_for_each_safe (pos, temp, &ahci.int_pending_list) {

		urb = list_entry (pos, urb_t, urb_list);
		
		assert (&urb->urb_list == pos->prev->next);
		assert (urb->hcpriv != 0);
		
        spin_unlock_irqrestore (&ahci.list_lock, flags);

		process_int_urb (urb);

		spin_lock_irqsave (&ahci.list_lock, flags);
	}
	
	spin_unlock_irqrestore (&ahci.list_lock, flags);
}

/* FIXME: TO DO 

static process_iso_list (void) {
	unsigned flags;
	
	struct list_head *pos;
	struct list_head *temp;
	urb_t * urb;
	
	spin_lock_irqsave (&ahci.list_lock, flags);
	
	list_for_each_safe (pos, temp, &ahci.iso_pending_list) {

		urb = list_entry (pos, urb_t, urb_list);
		
		assert (&urb->urb_list == pos->prev->next);
		assert (urb->hcpriv != 0);
		
        spin_unlock_irqrestore (&ahci.list_lock, flags);

		process_iso_urb (urb);
		
		spin_lock_irqsave (&ahci.list_lock, flags);
	}
	
	spin_unlock_irqrestore (&ahci.list_lock, flags);	
}

static void process_iso_urb (urb_t * urb) {
}

*/

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static void process_bulk_urb (urb_t * urb) {
	unsigned pipe = urb->pipe;
	unsigned dev = usb_pipedevice (pipe);
	unsigned ed = usb_pipeendpoint (pipe);
	unsigned maxpacket = usb_maxpacket (urb->dev, pipe, usb_pipeout (pipe));
	unsigned block_size = usb_pipeout(pipe)? AHCI_MAX_OUT_BLOCK_SIZE : AHCI_MAX_IN_BLOCK_SIZE;
	unsigned first_adr = 0;
	unsigned last_adr = 0;
	unsigned num_tds = 0;

	unsigned td_index;
	unsigned alloc_pos;
	unsigned restlen;
	
	struct ahci_urb_ext * pUrbExt = (struct ahci_urb_ext *)urb->hcpriv;
	struct ahci_endpoint * pahci_ep = ahci_ep_get (urb->dev, pipe);
	
							
	assert (usb_pipebulk(pipe));
	assert (pUrbExt != 0);

	info (pahci_ep != NULL);
	{	
		unsigned flags;

		spin_lock_irqsave (&ahci.list_lock, flags);
		if (pUrbExt->status || (pUrbExt->unlink_called && (pUrbExt->td_done == pUrbExt->tds_allocated))) {
			list_del (&urb->urb_list);
			
			if (!pUrbExt->status) {
				assert (!pUrbExt->unlink_called || (pUrbExt->td_done == pUrbExt->tds_allocated));
				pUrbExt->status = USB_ST_URB_KILLED;
			}
			if ((pahci_ep != NULL) && (pahci_ep->cur_urb == urb)) {
				pahci_ep->cur_urb = 0;
			}
			list_add_tail (&urb->urb_list, &ahci.completion_list);
			spin_unlock_irqrestore (&ahci.list_lock, flags);
			return;
		}
		spin_unlock_irqrestore (&ahci.list_lock, flags);
	}
	if (pahci_ep == NULL) {
		unlink_special (pUrbExt);
		return;
	}
	
	if (pahci_ep->cur_urb == 0) {
		pahci_ep->cur_urb = urb;
	}

	//info (pahci_ep->cur_urb == urb);
	if (pahci_ep->cur_urb != urb) {
			return;
	}
	
	assert (pUrbExt->tds_allocated < pUrbExt->td_count);
	
	td_index = pUrbExt->tds_allocated;
	alloc_pos = pUrbExt->alloc_pos;
	restlen = urb->transfer_buffer_length-alloc_pos;
	
	while (pahci_ep->pending_tds + num_tds < 4) {
		// Groe Blocks splitten
		unsigned sublen = AHCI_MIN (block_size, restlen);
		unsigned adr;

		assert (td_index < pUrbExt->td_count);

		adr = AHCI_MemoryAlloc (16 + (usb_pipeout(pipe)? sublen: 0));
		info (adr != 0);
		if (adr == 0) {
			break;
		}

		restlen -= sublen;

		if (num_tds > 0) {
			//tds verketten
			assert (last_adr != 0);
			assert (first_adr != 0);
			AHCI_PTR(last_adr)[3] = adr>>4;
		} else {
			assert (last_adr == 0);
			assert (first_adr == 0);
			first_adr = adr;
		}
		last_adr = adr;

		if (usb_pipeout(pipe)) {
			AHCI_InitBulkOutDesc (AHCI_PTR(adr), dev, ed, maxpacket, AHCI_ToggleTake, sublen, adr+16);
			AHCI_CopyToAlignedHW (AHCI_PTR(adr+16), ((char*)urb->transfer_buffer)+alloc_pos, sublen);
		} else {
			unsigned char bSPA = 1;
			AHCI_InitBulkInDesc (AHCI_PTR(adr), dev, ed, maxpacket, bSPA, AHCI_ToggleTake, sublen);
		}

		pUrbExt->td_adr[td_index] = adr;

		AHCI_MemorySetUrbPointer(adr, urb);
		
		alloc_pos += sublen;
		td_index++;
		num_tds++;
		
		if (restlen == 0) {
			assert (td_index == pUrbExt->td_count);
			break;
		}
	}
	assert (alloc_pos <= urb->transfer_buffer_length);
	assert (alloc_pos + restlen == urb->transfer_buffer_length);
	
	if (num_tds > 0) { 
		unsigned flags;

		spin_lock_irqsave (&ahci.list_lock, flags);
		if (!pUrbExt->unlink_called && !pUrbExt->status) {
			unsigned old_adr = pahci_ep->last_adr;
			assert (first_adr != 0);
			assert (last_adr != 0);
			
			pUrbExt->tds_allocated += num_tds;
			pUrbExt->alloc_pos = alloc_pos;

			pahci_ep->pending_tds += num_tds;
			pahci_ep->last_adr = last_adr;

			assert (td_index <= pUrbExt->td_count);
			if (td_index == pUrbExt->td_count) {
				list_del (&urb->urb_list);
				pahci_ep->cur_urb = 0;
			}

			//info (old_adr == 0);
			if (old_adr == 0) {
				AHCI_DWORD toggleFlag = usb_gettoggle(urb->dev, ed, usb_pipeout(urb->pipe)) << AHCI_DWORD2_initialToggle_SHIFT;
				
				AHCI_DWORD tmp = AHCI_PTR(first_adr)[1] & ~(AHCI_DWORD2_initialToggle_FLAG | AHCI_DWORD2_takeToggle_FLAG);
				AHCI_PTR(first_adr)[1] = (toggleFlag | tmp);
				
				AHCI_InsertDescriptor (AHCI_LIST_lowPrio, first_adr);
			} else {
				AHCI_AppendDescriptor (AHCI_LIST_lowPrio, old_adr, first_adr);
			}
			spin_unlock_irqrestore (&ahci.list_lock, flags);
		} else {
			spin_unlock_irqrestore (&ahci.list_lock, flags);
			while (num_tds > 0) {
				num_tds --;
				AHCI_MemoryFree (pUrbExt->td_adr[pUrbExt->tds_allocated+num_tds]);
			}
		}
		spin_unlock_irqrestore (&ahci.list_lock, flags);
	}

	//PrintInfo ("Bulk Descriptor @0x%04X: %08X %08X %08X %08X\n",
	//				adr,AHCI_PTR(adr)[0], AHCI_PTR(adr)[1], AHCI_PTR(adr)[2], AHCI_PTR(adr)[3]);

}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static void process_control_urb (urb_t * urb) {
	unsigned pipe = urb->pipe;
	unsigned dev = usb_pipedevice (pipe);
	unsigned ed = usb_pipeendpoint (pipe);
	unsigned maxpacket = usb_maxpacket (urb->dev, pipe, usb_pipeout (pipe));
	unsigned block_size = usb_pipeout(pipe)? AHCI_MAX_OUT_BLOCK_SIZE : AHCI_MAX_IN_BLOCK_SIZE;
	unsigned first_adr = 0;
	unsigned last_adr = 0;
	unsigned num_tds = 0;

	unsigned td_index;
	unsigned alloc_pos;
	unsigned restlen;
	
	struct ahci_urb_ext * pUrbExt = (struct ahci_urb_ext *)urb->hcpriv;
	struct ahci_endpoint * pahci_ep = ahci_ep_get (urb->dev, pipe);
							
	assert (usb_pipecontrol(pipe));
	assert (pUrbExt != 0);
	
	info (pahci_ep != NULL);
	{	
		unsigned flags;

		spin_lock_irqsave (&ahci.list_lock, flags);
		if (pUrbExt->status || (pUrbExt->unlink_called && (pUrbExt->td_done == pUrbExt->tds_allocated))) {
			list_del (&urb->urb_list);
			
			if (!pUrbExt->status) {
				assert (!pUrbExt->unlink_called || (pUrbExt->td_done == pUrbExt->tds_allocated));
				pUrbExt->status = USB_ST_URB_KILLED;
			}
			if ((pahci_ep != NULL) && (pahci_ep->cur_urb == urb)) {
				pahci_ep->cur_urb = 0;
			}
			list_add_tail (&urb->urb_list, &ahci.completion_list);
			spin_unlock_irqrestore (&ahci.list_lock, flags);
			return;
		}
		spin_unlock_irqrestore (&ahci.list_lock, flags);
	}
	if (pahci_ep == NULL) {
		unlink_special (pUrbExt);
		return;
	}
	
	if (pahci_ep->cur_urb == 0) {
		pahci_ep->cur_urb = urb;
	}

	info (pahci_ep->cur_urb == urb);
	if (pahci_ep->cur_urb != urb) {
			return;
	}
	
	assert (pUrbExt->tds_allocated < pUrbExt->td_count);
	
	td_index = pUrbExt->tds_allocated;
	alloc_pos = pUrbExt->alloc_pos;
	restlen = urb->transfer_buffer_length-alloc_pos;

	//PrintInfo ("proc control: urb %p td_index %u , restlen = %u , td_count %u\n", urb, td_index, restlen, pUrbExt->td_count);
	
	while (pahci_ep->pending_tds + num_tds < 3) {
		unsigned adr;
		if (td_index == 0) {
			adr = AHCI_MemoryAlloc (16+8);
			info (adr != 0);
			if (adr == 0) {
				break;
			}

			AHCI_InitSetupDesc (AHCI_PTR(adr), dev, ed, adr+16);
			AHCI_DWORDCopyToAlignedHW (AHCI_PTR (adr+16), urb->setup_packet, 8);
			first_adr = adr;
			last_adr = adr;
			//PrintInfo ("Control Setup Descriptor @0x%04X: %08X %08X %08X %08X urb%p\n",
			//			adr, AHCI_PTR(adr)[0], AHCI_PTR(adr)[1], AHCI_PTR(adr)[2], AHCI_PTR(adr)[3], urb);

			AHCI_MemorySetUrbPointer(adr, urb);
			pUrbExt->td_adr[0] = adr;
			td_index ++;
			num_tds ++;
		} else if (td_index == pUrbExt->td_count-1) {
			adr = AHCI_MemoryAlloc (16);
			info (adr != 0);
			if (adr == 0) {
				break;
			}
			if (num_tds > 0) {
				//tds verketten
				assert (last_adr != 0);
				assert (first_adr != 0);
				AHCI_PTR(last_adr)[3] = adr>>4;
			} else {
				assert (last_adr == 0);
				assert (first_adr == 0);
				first_adr = adr;
			}

			if (usb_pipein(pipe)) {
				AHCI_InitBulkOutDesc (AHCI_PTR(adr), dev, ed, 0, AHCI_Toggle1, 0, 0);
			} else {
				// no short packet
				AHCI_InitBulkInDesc (AHCI_PTR(adr), dev, ed, maxpacket, 1, AHCI_Toggle1, 0);
			}
			last_adr = adr;
			
			AHCI_MemorySetUrbPointer(adr, urb);
			pUrbExt->td_adr[td_index] = adr;
			td_index ++;
			num_tds ++;
			assert (td_index == pUrbExt->td_count);
			//PrintInfo ("Control Handshake Descriptor @0x%04X: %08X %08X %08X %08X\n",
			//			adr,AHCI_PTR(adr)[0], AHCI_PTR(adr)[1], AHCI_PTR(adr)[2], AHCI_PTR(adr)[3]);
			break;
			
		} else {
			// Groe Blocks splitten
			unsigned sublen = AHCI_MIN (block_size, restlen);
			assert (urb->transfer_buffer_length > 0);
			assert (restlen > 0);

			assert (td_index < pUrbExt->td_count-1);

			adr = AHCI_MemoryAlloc (16 + (usb_pipeout(pipe)? sublen: 0));

			info (adr != 0);
			if (adr == 0) {
				break;
			}
			restlen -= sublen;

			if (num_tds > 0) {
				//tds verketten
				assert (last_adr != 0);
				assert (first_adr != 0);
				AHCI_PTR(last_adr)[3] = adr>>4;
			} else {
				assert (last_adr == 0);
				assert (first_adr == 0);
				first_adr = adr;
			}
			last_adr = adr;

			if (usb_pipeout(pipe)) {
				AHCI_InitBulkOutDesc (AHCI_PTR(adr), dev, ed, maxpacket, AHCI_ToggleTake, sublen, adr+16);
				AHCI_CopyToAlignedHW (AHCI_PTR(adr+16), ((char*)urb->transfer_buffer)+alloc_pos, sublen);
			} else {
				unsigned char bSPA = 1;
				AHCI_InitBulkInDesc (AHCI_PTR(adr), dev, ed, maxpacket, bSPA, AHCI_ToggleTake, sublen);
			}

			pUrbExt->td_adr[td_index] = adr;

			AHCI_MemorySetUrbPointer(adr, urb);

			alloc_pos += sublen;
			td_index++;
			num_tds++;
			//PrintInfo ("Control Data Descriptor @0x%04X: %08X %08X %08X %08X\n",
			//			adr,AHCI_PTR(adr)[0], AHCI_PTR(adr)[1], AHCI_PTR(adr)[2], AHCI_PTR(adr)[3]);
		}
	}
	assert (alloc_pos <= urb->transfer_buffer_length);
	assert (alloc_pos + restlen == urb->transfer_buffer_length);
	
	if (num_tds > 0) { 
		unsigned flags;

		spin_lock_irqsave (&ahci.list_lock, flags);
		if (!pUrbExt->unlink_called && !pUrbExt->status) {
			unsigned old_adr = pahci_ep->last_adr;
			assert (first_adr != 0);
			assert (last_adr != 0);
			
			pUrbExt->tds_allocated += num_tds;
			pUrbExt->alloc_pos = alloc_pos;

			pahci_ep->pending_tds += num_tds;
			pahci_ep->last_adr = last_adr;

			assert (td_index <= pUrbExt->td_count);
			if (td_index == pUrbExt->td_count) {
				list_del (&urb->urb_list);
				pahci_ep->cur_urb = 0;
			}

			//info (old_adr == 0);
			if (old_adr == 0) {
				AHCI_InsertDescriptor (AHCI_LIST_lowPrio, first_adr);
			} else {
				AHCI_AppendDescriptor (AHCI_LIST_lowPrio, old_adr, first_adr);
			}
			spin_unlock_irqrestore (&ahci.list_lock, flags);
		} else {
			spin_unlock_irqrestore (&ahci.list_lock, flags);
			while (num_tds > 0) {
				num_tds --;
				AHCI_MemoryFree (pUrbExt->td_adr[pUrbExt->tds_allocated+num_tds]);
			}
		}
		spin_unlock_irqrestore (&ahci.list_lock, flags);
	}
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
static void process_int_urb (urb_t * urb) {
	unsigned pipe = urb->pipe;
	unsigned dev = usb_pipedevice (pipe);
	unsigned ed = usb_pipeendpoint (pipe);
	unsigned maxpacket = usb_maxpacket (urb->dev, pipe, usb_pipeout (pipe));
	unsigned first_adr = 0;
	unsigned last_adr = 0;
	unsigned num_tds = 0;

	unsigned td_index;
	unsigned alloc_pos;
	unsigned restlen;
	
	struct ahci_urb_ext * pUrbExt = (struct ahci_urb_ext *)urb->hcpriv;
	struct ahci_endpoint * pahci_ep = ahci_ep_get (urb->dev, pipe);
							
	assert (usb_pipeint(pipe));
	assert (usb_pipein(pipe));
	assert (pUrbExt != 0);
	
	info (pahci_ep != NULL);
	{	
		unsigned flags;

		spin_lock_irqsave (&ahci.list_lock, flags);
		if (pUrbExt->status || (pUrbExt->unlink_called && (pUrbExt->td_done == pUrbExt->tds_allocated))) {
			list_del (&urb->urb_list);
			
			if (!pUrbExt->status) {
				assert (!pUrbExt->unlink_called || (pUrbExt->td_done == pUrbExt->tds_allocated));
				pUrbExt->status = USB_ST_URB_KILLED;
			}
			if ((pahci_ep != NULL) && (pahci_ep->cur_urb == urb)) {
				pahci_ep->cur_urb = 0;
			}
			list_add_tail (&urb->urb_list, &ahci.completion_list);
			spin_unlock_irqrestore (&ahci.list_lock, flags);
			return;
		}
		spin_unlock_irqrestore (&ahci.list_lock, flags);
	}
	if (pahci_ep == NULL) {
		unlink_special (pUrbExt);
		return;
	}
	if (pahci_ep->cur_urb == 0) {
		pahci_ep->cur_urb = urb;
	}

	info (pahci_ep->cur_urb == urb);
	if (pahci_ep->cur_urb != urb) {
			return;
	}
	
	assert (pUrbExt->tds_allocated < pUrbExt->td_count);
	
	td_index = pUrbExt->tds_allocated;
	alloc_pos = pUrbExt->alloc_pos;
	restlen = urb->transfer_buffer_length-alloc_pos;
	
	while (pahci_ep->pending_tds + num_tds < 3) {
		// Groe Blocks splitten
		unsigned sublen = AHCI_MIN (AHCI_MAX_IN_BLOCK_SIZE, restlen);
		unsigned adr;

		assert (td_index < pUrbExt->td_count);

		adr = AHCI_MemoryAlloc (16 + sublen);
		info (adr != 0);
		if (adr == 0) {
			break;
		}
		restlen -= sublen;

		if (num_tds > 0) {
			//tds verketten
			assert (last_adr != 0);
			assert (first_adr != 0);
			AHCI_PTR(last_adr)[3] = adr>>4;
		} else {
			assert (last_adr == 0);
			assert (first_adr == 0);
			first_adr = adr;
		}

		last_adr = adr;
		{
			unsigned char bSPA = 1;
			unsigned start_frame = (AHCI_GetFrameNumber () + urb->interval +10 ) & 0xFFFF;
			if (td_index == 0) {
				urb->start_frame = start_frame;
			}
 			AHCI_InitIntDesc (AHCI_PTR(adr), start_frame, urb->interval, 
							dev, ed, maxpacket, bSPA, AHCI_ToggleTake, sublen);
		}

		pUrbExt->td_adr[td_index] = adr;

		AHCI_MemorySetUrbPointer(adr, urb);
		
		alloc_pos += sublen;
		td_index++;
		num_tds++;
		
		if (restlen == 0) {
			break;
		}
	}
	assert (alloc_pos <= urb->transfer_buffer_length);
	assert (alloc_pos + restlen == urb->transfer_buffer_length);
	
	if (num_tds > 0) { 
		unsigned flags;

		spin_lock_irqsave (&ahci.list_lock, flags);
		if (!pUrbExt->unlink_called && !pUrbExt->status) {
			unsigned old_adr = pahci_ep->last_adr;
			assert (first_adr != 0);
			assert (last_adr != 0);
			
			pUrbExt->tds_allocated += num_tds;
			pUrbExt->alloc_pos = alloc_pos;

			pahci_ep->pending_tds += num_tds;
			pahci_ep->last_adr = last_adr;

			assert (td_index <= pUrbExt->td_count);
			if (td_index == pUrbExt->td_count) {
				list_del (&urb->urb_list);
				pahci_ep->cur_urb = 0;
			}

			//info (old_adr == 0);
			if (old_adr == 0) {
				AHCI_DWORD toggleFlag = usb_gettoggle(urb->dev, ed, usb_pipeout(urb->pipe)) << AHCI_DWORD2_initialToggle_SHIFT;
				
				AHCI_DWORD tmp = AHCI_PTR(first_adr)[1] & ~(AHCI_DWORD2_initialToggle_FLAG | AHCI_DWORD2_takeToggle_FLAG);
				AHCI_PTR(first_adr)[1] = (toggleFlag | tmp);
				
				AHCI_InsertDescriptor (AHCI_LIST_highPrio, first_adr);
			} else {
				AHCI_AppendDescriptor (AHCI_LIST_highPrio, old_adr, first_adr);
			}
			spin_unlock_irqrestore (&ahci.list_lock, flags);
		} else {
			spin_unlock_irqrestore (&ahci.list_lock, flags);
			while (num_tds > 0) {
				num_tds --;
				AHCI_MemoryFree (pUrbExt->td_adr[pUrbExt->tds_allocated+num_tds]);
			}
		}
		spin_unlock_irqrestore (&ahci.list_lock, flags);
	}

	//PrintInfo ("Int Descriptor @0x%04X: %08X %08X %08X %08X\n",
	//				adr,AHCI_PTR(adr)[0], AHCI_PTR(adr)[1], AHCI_PTR(adr)[2], AHCI_PTR(adr)[3]);

}

/*---------------------------------------------------------------------------*\
	callback for tasklet
\*---------------------------------------------------------------------------*/
void urb_completion_func (unsigned long data) {
	
	unsigned long flags;
	
	process_control_list ();
	//process_iso_list ();
	process_int_list ();
	process_bulk_list ();
	
	//PrintInfo ("urb_completion_func enter\n");
	spin_lock_irqsave (&ahci.list_lock, flags);
	
	while (!list_empty (&ahci.completion_list)) {
		urb_t * urb = list_entry (ahci.completion_list.next, urb_t, urb_list);
		assert (&urb->urb_list == ahci.completion_list.next);
		assert (urb->hcpriv != 0);
		
		list_del (&urb->urb_list);
		spin_unlock_irqrestore (&ahci.list_lock, flags);
		{
			struct ahci_urb_ext * pUrbExt = urb->hcpriv;
			unsigned unlink_called = pUrbExt->unlink_called;

			int status = pUrbExt->status;

			assert (unlink_called || status != USB_ST_URB_KILLED);
			if ((status == USB_ST_URB_KILLED) && (urb->transfer_flags & URB_ASYNC_UNLINK)) {
				status = -ECONNRESET;
			}

			assert (urb->status == USB_ST_URB_PENDING);
			urb->status = status; 
			assert (urb->status != USB_ST_URB_PENDING);

			if (unlink_called) {
				wait_queue_head_t * wait_head = pUrbExt->wait_head;
				PrintInfo ("wake urb %p wh=%p\n", urb, wait_head);
				if (wait_head) {
					assert (!(urb->transfer_flags & URB_ASYNC_UNLINK));
					wake_up (wait_head);
				}
			}
			ahci_free_urb_ext (urb);

			if (status != 0) {
				PrintInfo ("complete urb %p with st %d\n", urb, status);
			}
			
			if (urb->complete) {
				//PrintInfo ("complete urb %p\n", urb);
				urb->complete (urb, &dummy_regs);
			}
		}
		spin_lock_irqsave (&ahci.list_lock, flags);
	}
	spin_unlock_irqrestore (&ahci.list_lock, flags);
	//PrintInfo ("urb_completion_func leave\n");
}
