/*
 * drivers/video/dm64xxfb.c
 *
 * Framebuffer driver for Texas Instruments DM644x display controller.
 *
 * Copyright (C) 2005 Texas Instruments, Inc.
 * Rishi Bhattacharya <support@ti.com>
 *
 * This file is licensed under the terms of the GNU General Public License
 * version 2. This program is licensed "as is" without any warranty of any
 * kind, whether express or implied.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/dma-mapping.h>
#include <linux/interrupt.h>
#include <asm/irq.h>

#include "davincifb.h"

/* #define	DEBUG */
#ifdef 	DEBUG
#define	DBGENTER	printk("Entered %s\n", __FUNCTION__)
#define	DBGEXIT		printk("Exited %s\n", __FUNCTION__)
#define DBGPRINT(x...)  printk(x)
#define	RETURN(x)						\
do {								\
	int __ret = (x);					\
	printk("Exited %s : ret %d\n", __FUNCTION__, __ret);	\
	return __ret; 						\
} while(0)
#else
#define	DBGENTER
#define	DBGEXIT
#define DBGPRINT(x...)
#define	RETURN(x)	return (x)
#endif

/*
 * display controller register I/O routines
 */
static __inline__ u32
dispc_reg_in(u32 offset)
{
	return(inl(offset));
}
static __inline__ u32
dispc_reg_out(u32 offset, u32 val)
{
	/* outw(val, offset);  */
	outl(val, offset);
	return(val);
}
static __inline__ u32
dispc_reg_merge(u32 offset, u32 val, u32 mask)
{
	u32 addr = offset;
	u32 new_val = (inl(addr) & ~mask) | (val & mask);
	outl(new_val, addr);
	return(new_val);
}

static void venc_config(void);
static void venc_start(void);

/* There are 4 framebuffers, each represented by an fb_info and
 * a dm_win_info structure
 */

#define DRIVER		"dm64xxfb"
#define OSD0_FBNAME	"dm_osd0_fb"
#define OSD1_FBNAME	"dm_osd1_fb"
#define VID0_FBNAME	"dm_vid0_fb"
#define VID1_FBNAME	"dm_vid1_fb"

/* usage:	if (is_win(info->fix.id, OSD0)) ... */
#define is_win(name, x) ((strcmp(name, x ## _FBNAME) == 0) ? 1 : 0)

struct dm_win_info {
	struct fb_info	info;

	/* framebuffer area */
	dma_addr_t fb_base_phys;
	unsigned long fb_base;
	unsigned long fb_size;

	u32 pseudo_palette[17];

	/* flag to identify if framebuffer area is fixed already or not */
	int alloc_fb_mem;

	struct dm_info *dm;
};

static struct dm_info {
	struct dm_win_info *osd0;
	struct dm_win_info *osd1;
	struct dm_win_info *vid0;
	struct dm_win_info *vid1;

	/* to map the registers */
	dma_addr_t mmio_base_phys;
	unsigned long mmio_base;
	unsigned long mmio_size;

	wait_queue_head_t vsync_wait;
	unsigned long vsync_cnt;
	int timeout;
} dm;

static struct fb_ops dm64xxfb_ops;

#define BASEX		0x80
#define BASEY		0x12

/* Random value chosen for now. Should be within the panel's supported range */
#define LCD_PANEL_CLOCK	180000

#define DISP_XRES	720
#define DISP_YRES	480

/* All window widths have to be rounded up to a multiple of 32 bytes */

/* The OSD0 window has to be always within VID0. Plus, since it is in RGB565
 * mode, it _cannot_ overlap with VID1.
 * be displayed in the top left quadrant of the screen, and the VID1 in the
 * bottom right quadrant. So the default 'xres' and 'yres' are set to  half of
 * the screen width and height respectively. Note however that the framebuffer
 * size is allocated for the full screen size so the user can change the 'xres'
 * and 'yres' by using the FBIOPUT_VSCREENINFO ioctl within the limits of the
 * screen size.
 */
#define round_32(width)	((((width) + 31) / 32) * 32 )

//#define OSD0_XRES	((((DISP_XRES/2 * 16/8) + 31) / 32) * 32 * 8/16)
#define OSD0_XRES	round_32((DISP_XRES)*16/8) * 8/16	/* pixels */
			/* this is to round the xres in 'bytes' to a multiple
			 * of 32 bytes
			 */
#define OSD0_YRES	DISP_YRES
#define OSD0_FB_PHY	0
//#define OSD0_FB_SIZE	(((((DISP_XRES * 16/8) + 31) / 32) * 32 * 8/16) * DISP_YRES * 16/8 * 2)
#define OSD0_FB_SIZE	(round_32((DISP_XRES)*16/8) * DISP_YRES * 2)
			/* 16 bpp, Double buffered */
static struct fb_var_screeninfo osd0_default_var = {
	.xres		= OSD0_XRES,
	.yres		= OSD0_YRES,
	.xres_virtual	= OSD0_XRES,
	.yres_virtual	= OSD0_YRES*2,
	.xoffset	= 0,
	.yoffset	= 0,
	.bits_per_pixel	= 16,
	.grayscale	= 0,
	.red		= {11, 5, 0},
	.green		= { 5, 6, 0},
	.blue		= { 0, 5, 0},
	.transp		= { 0, 0, 0},
	.nonstd		= 0,
	.activate	= FB_ACTIVATE_NOW,
	.height		= -1,
	.width		= -1,
	.accel_flags	= 0,
	.pixclock	= LCD_PANEL_CLOCK, /* picoseconds */
	.left_margin	= 40,		/* pixclocks */
	.right_margin	= 4,		/* pixclocks */
	.upper_margin	= 8,		/* line clocks */
	.lower_margin	= 2,		/* line clocks */
	.hsync_len	= 4,		/* pixclocks */
	.vsync_len	= 2,		/* line clocks */
	.sync		= 0,
	.vmode		= FB_VMODE_INTERLACED,
	.reserved[0]	= 0,		/* X position */
	.reserved[1]	= 0,		/* Y position */
	.reserved[3]	= 0,		/* Zoom */
};

/* Using the full screen for OSD1 by default */
//#define OSD1_XRES	((((DISP_XRES * 4/8) + 31) / 32) * 32 * 8/4)
#define OSD1_XRES	round_32(DISP_XRES*4/8) * 8/4	/* pixels */
#define OSD1_YRES	DISP_YRES
#define OSD1_FB_PHY	0
//#define OSD1_FB_SIZE	(((((DISP_XRES * 4/8) + 31) / 32) * 32 * 8/4) * DISP_YRES * 4/8 * 2)
#define OSD1_FB_SIZE	(round_32(DISP_XRES*4/8) * DISP_YRES * 2)

static struct fb_var_screeninfo osd1_default_var = {
	.xres		= OSD1_XRES,
	.yres		= OSD1_YRES,
	.xres_virtual	= OSD1_XRES,
	.yres_virtual	= OSD1_YRES*2,
	.xoffset	= 0,
	.yoffset	= 0,
	.bits_per_pixel	= 4,
	.activate	= FB_ACTIVATE_NOW,
	.accel_flags	= 0,
	.pixclock	= LCD_PANEL_CLOCK, /* picoseconds */
	.vmode		= FB_VMODE_INTERLACED,
	.reserved[0]	= 0,		/* X position */
	.reserved[1]	= 0,		/* Y position */
	.reserved[3]	= 0,		/* Zoom */
};

/* Using the full screen for OSD0 by default */
//#define VID0_XRES	((((DISP_XRES * 16/8) + 31) / 32) * 32 * 8/16)
// #define VID0_XRES	round_32(DISP_XRES*16/8)	/* pixels */
#define VID0_XRES		round_32(DISP_XRES*16/8) * 8/16	/* pixels */
#define VID0_YRES	DISP_YRES
#define VID0_FB_PHY	0
//#define VID0_FB_SIZE	(((((DISP_XRES * 16/8) + 31) / 32) * 32 * 8/16) * DISP_YRES * 16/8 * 2)
#define VID0_FB_SIZE	(round_32(DISP_XRES*16/8) * DISP_YRES * 3)
static struct fb_var_screeninfo vid0_default_var = {
	.xres		= VID0_XRES,
	.yres		= VID0_YRES,
	.xres_virtual	= VID0_XRES,
	.yres_virtual	= VID0_YRES*3,
	.xoffset	= 0,
	.yoffset	= 0,
	.bits_per_pixel	= 16,
	.activate	= FB_ACTIVATE_NOW,
	.accel_flags	= 0,
	.pixclock	= LCD_PANEL_CLOCK, /* picoseconds */
	.vmode		= FB_VMODE_INTERLACED,
	.reserved[0]	= 0,		/* X position */
	.reserved[1]	= 0,		/* Y position */
	.reserved[3]	= 0,		/* Zoom */
};

/* Using the bottom right quadrant of the screen screen for VID1 by default,
 * but keeping the framebuffer allocated for the full screen, so the user can
 * change the 'xres' and 'yres' later using the FBIOPUT_VSCREENINFO ioctl.
 * */
//#define VID1_XRES	(DISP_XRES - OSD0_XRES)
#define VID1_XRES	round_32((DISP_XRES - OSD0_XRES)*16/8)
#define VID1_YRES	(DISP_YRES - OSD0_YRES)
#define VID1_FB_PHY	0
#define VID1_FB_SIZE	(round_32(DISP_XRES*16/8) * DISP_YRES * 3)
static struct fb_var_screeninfo vid1_default_var = {
	.xres		= VID1_XRES,
	.yres		= VID1_YRES,
	.xres_virtual	= VID1_XRES,
	.yres_virtual	= VID1_YRES*3,
	.xoffset	= 0,
	.yoffset	= 0,
	.bits_per_pixel	= 16,
	.activate	= FB_ACTIVATE_NOW,
	.accel_flags	= 0,
	.pixclock	= LCD_PANEL_CLOCK, /* picoseconds */
	.vmode		= FB_VMODE_INTERLACED,
	.reserved[0]	= 0,		/* X position */
	.reserved[1]	= 0,		/* Y position */
	.reserved[3]	= 0,		/* Zoom */
};

#define	x_pos(v)	((v)->reserved[0])
#define	y_pos(v)	((v)->reserved[1])
#define	zoom(v)		((v)->reserved[2])

/* Must do checks against the limits of the output device */
static int dm64xx_venc_check_mode(const struct dm_win_info *w,
				  const struct fb_var_screeninfo *var)
{
	DBGENTER;
	RETURN(0);
}

static irqreturn_t dm64xxfb_isr(int irq, void *arg, struct pt_regs *regs)
{
	struct dm_info *dm = (struct dm_info *) arg;
	++dm->vsync_cnt;
	wake_up_interruptible(&dm->vsync_wait);
	return IRQ_HANDLED;
}

/* Wait for a vsync interrupt.  This routine sleeps so it can only be called
 * from process context.
 */
static int dm64xxfb_wait_for_vsync(struct dm_win_info *w)
{
	struct dm_info *dm = w->dm;
	wait_queue_t wq;
	unsigned long cnt;
	int ret;

	DBGENTER;
	init_waitqueue_entry(&wq, current);

	cnt = dm->vsync_cnt;
	ret = wait_event_interruptible_timeout(dm->vsync_wait,
		cnt != dm->vsync_cnt, dm->timeout);
	if (ret < 0)
		RETURN(ret);
	if (ret == 0)
		RETURN(-ETIMEDOUT);

	RETURN(0);
}

/* These position parameters are given through fb_var_screeninfo.
 * xp = var.reserved[0], yp = var.reserved[1],
 * xl = var.xres, yl = var.yres
 */
static void set_win_position(char *id, u32 xp, u32 yp, u32 xl, u32 yl)
{
	int i = 0;

	DBGENTER;
	if (is_win(id, VID0))  {
		i = 0;
	DBGPRINT("Video Window 0 X Pixel Offset =%d\n", xp);
	DBGPRINT("Video Window 0 Y Pixel Offset =%d\n", yp);
	DBGPRINT("Video Window 0 X length =%d\n", xl);
	DBGPRINT("Video Window 0 Y length =%d\n", yl);
	DBGPRINT("VID0 XRES = =0x%p\n", VID0_XRES);
	}

	else if (is_win(id, VID1))
		i = 1;
	else if (is_win(id, OSD0))  {
		i = 2;
	DBGPRINT("OSD Window 0 X Pixel Offset =%d\n", xp);
	DBGPRINT("OSD Window 0 Y Pixel Offset =%d\n", yp);
	DBGPRINT("OSD Window 0 X length =%d\n", xl);
	DBGPRINT("OSD Window 0 Y length =%d\n", yl);

	}

	else if (is_win(id, OSD1))  {
		i = 3;
	DBGPRINT("OSD Window 1 X Pixel Offset =%d\n", xp);
	DBGPRINT("OSD Window 1 Y Pixel Offset =%d\n", yp);
	DBGPRINT("OSD Window 1 X length =%d\n", xl);
	DBGPRINT("OSD Window 1 Y length =%d\n", yl);

	}

	dispc_reg_out(OSD_WINXP(i), xp);
	dispc_reg_out(OSD_WINYP(i), yp);
	dispc_reg_out(OSD_WINXL(i), xl);
	dispc_reg_out(OSD_WINYL(i), yl);
	DBGEXIT;
}
/*
static void get_win_position(char *id, u32 *xp, u32 *yp, u32 *xl, u32 *yl)
{
	int i = 0;

	DBGENTER;
	if (is_win(id, VID0))
		i = 0;
	else if (is_win(id, VID1))
		i = 1;
	else if (is_win(id, OSD0))
		i = 2;
	else if (is_win(id, OSD1))
		i = 3;

	*xp = dispc_reg_in(OSD_WINXP(i)) & OSD_VIDWIN0XP_V0X;
	*yp = dispc_reg_in(OSD_WINYP(i)) & OSD_VIDWIN0YP_V0Y;
	*xl = dispc_reg_in(OSD_WINXL(i)) & OSD_VIDWIN0XL_V0W;
	*yl = dispc_reg_in(OSD_WINYL(i)) & OSD_VIDWIN0YL_V0H;
	DBGEXIT;
}
*/

static inline void get_win_position(struct dm_win_info *w,
				u32 *xp, u32 *yp, u32 *xl, u32 *yl)
{
	struct fb_var_screeninfo *v = &w->info.var;
	*xp = x_pos(v);
	*yp = y_pos(v);
	*xl = v->xres;
	*yl = v->yres;
}

#define OVERLAP(x1, y1, x2, y2, x3, y3, x4, y4)		\
(!(	((x1)<(x3) && (x2)<(x3)) || ((x1)>(x4) && (x2)>(x4)) ||	\
	((y1)<(y3) && (y2)<(y3)) || ((y1)>(y4) && (y2)>(y4)) )	\
)

/* Returns 1 if the windows overlap, 0 otherwise */
static int window_overlap(struct dm_win_info *w,
			  u32 xp, u32 yp, u32 xl, u32 yl)
{
	u32 _xp = 0, _yp = 0, _xl = 0, _yl = 0;

	DBGENTER;
	if (!w)
		RETURN(0);

	get_win_position(w, &_xp, &_yp, &_xl, &_yl);
	DBGPRINT("_xp %d, _yp %d, _xl %d, _yl %d\n", _xp, _yp, _xl, _yl);

	RETURN(OVERLAP(xp, yp, xp + xl, yp + yl,
			_xp, _yp, _xp + _xl, _xp + _xl));
}

/* Returns 1 if the window parameters are within VID0, 0 otherwise */
static int within_vid0_limits(u32 xp, u32 yp, u32 xl, u32 yl)
{
	u32 vid0_xp = 0, vid0_yp = 0, vid0_xl = 0, vid0_yl = 0;

	DBGENTER;
	if (!dm.vid0)
		RETURN(1);

	get_win_position(dm.vid0, &vid0_xp, &vid0_yp, &vid0_xl, &vid0_yl);
	DBGPRINT("vid0 : xp %d, yp %d, xl %d, yl %d\n",
		vid0_xp, vid0_yp, vid0_xl, vid0_yl);

	if ((xp >= vid0_xp) && (yp >= vid0_yp) &&
	    (xl <= vid0_xl) && (yl <= vid0_yl))
		RETURN(1);

	RETURN(0);
}

/* VID0 must be large enough to hold all other windows */
static int check_new_vid0_size(u32 xp0, u32 yp0, u32 xl0, u32 yl0)
{
	u32 _xp = 0, _yp = 0, _xl = 0, _yl = 0;

	if (dm.osd0) {
		get_win_position(dm.osd0, &_xp, &_yp, &_xl, &_yl);
		if (!((_xp >= xp0) && (_yp >= yp0) &&
		      (_xl <= xl0) && (_yl <= yl0)))
			RETURN(-EINVAL);
	}
	if (dm.osd1) {
		get_win_position(dm.osd1, &_xp, &_yp, &_xl, &_yl);
		if (!((_xp >= xp0) && (_yp >= yp0) &&
		      (_xl <= xl0) && (_yl <= yl0)))
			RETURN(-EINVAL);
	}
	if (dm.vid1) {
		get_win_position(dm.vid1, &_xp, &_yp, &_xl, &_yl);
		if (!((_xp >= xp0) && (_yp >= yp0) &&
		      (_xl <= xl0) && (_yl <= yl0)))
			RETURN(-EINVAL);
	}
	RETURN(0);
}

/*---------------------------------------------------------------------------*/

/**
 *      dm64xxfb_check_var - Optional function. Validates a var passed in.
 *      @var: frame buffer variable screen structure
 *      @info: frame buffer structure that represents a single frame buffer
 *
 *	Checks to see if the hardware supports the state requested by
 *	var passed in. This function does not alter the hardware state!!!
 *	This means the data stored in struct fb_info and struct xxx_par do
 *      not change. This includes the var inside of struct fb_info.
 *	Do NOT change these. This function can be called on its own if we
 *	intent to only test a mode and not actually set it. The stuff in
 *	modedb.c is a example of this. If the var passed in is slightly
 *	off by what the hardware can support then we alter the var PASSED in
 *	to what we can do. If the hardware doesn't support mode change
 * 	a -EINVAL will be returned by the upper layers. You don't need to
 *	implement this function then. If you hardware doesn't support
 *	changing the resolution then this function is not needed. In this
 *	case the driver would just provide a var that represents the static
 *	state the screen is in.
 *
 *	Returns negative errno on error, or zero on success.
 */
static int dm64xxfb_check_var(struct fb_var_screeninfo *var,
			      struct fb_info *info)
{
	const struct dm_win_info *w = (const struct dm_win_info *) info->par;
	struct fb_var_screeninfo v;


/* Rules:
 * 1) Vid1, OSD0, OSD1 and Cursor must be fully contained inside of Vid0.
 * 2) Vid0 and Vid1 are both set to accept YUV 4:2:2 (for now).
 * 3) OSD window data is always packed into 32-bit words and left justified.
 * 4) Each horizontal line of window data must be a multiple of 32 bytes.
 *    32 bytes = 32 bytes / 2 bytes per pixel = 16 pixels.
 *    This implies that 'xres' must be a multiple of 32 bytes.
 * 5) The offset registers hold the distance between the start of one line and
 *    the start of the next. This offset value must be a multiple of 32 bytes.
 *    This implies that 'xres_virtual' is also a multiple of 32 bytes. Note
 *    that 'xoffset' needn't be a multiple of 32 bytes.
 * 6) OSD0 is set to accept RGB565.
 * 	dispc_reg_merge(OSD_OSDWIN0ND, OSD_OSDWIN0ND_RGB0E, OSD_OSDWIN0ND_RGB0E)
 * 7) OSD1 is set to be the attribute window.
 * 8) Vid1 startX = Vid0 startX + N * 16 pixels (32 bytes)
 * 9) Vid1 width = (16*N - 8) pixels
 * 10) When one of the OSD windows is in RGB565, it cannot overlap with Vid1.
 */

	DBGENTER;
	memcpy(&v, var, sizeof(v));
	RETURN(0);
	/* do board-specific checks on the var */
	if (dm64xx_venc_check_mode(w, &v))
		RETURN(-EINVAL);

	if (v.xres_virtual < v.xres || v.yres_virtual < v.yres)
		RETURN(-EINVAL);
	if (v.xoffset > v.xres_virtual - v.xres)
		RETURN(-EINVAL);
	if (v.yoffset > v.yres_virtual - v.yres)
		RETURN(-EINVAL);
	if ((v.xres * v.bits_per_pixel/8) % 32 ||
	    (v.xres_virtual * v.bits_per_pixel/8) % 32)	/* Rules 4, 5 */
		RETURN(-EINVAL);
	if (v.xres_virtual * v.yres_virtual > w->fb_size)
		RETURN(-EINVAL);

	if (!is_win(info->fix.id, VID0)) {
		/* Rule 1 */
		if (!within_vid0_limits(x_pos(&v), y_pos(&v),
					v.xres_virtual - v.xres,
					v.yres_virtual - v.yres))
			RETURN(-EINVAL);
	}
	if (is_win(info->fix.id, OSD0)) {
		/* Rule 10 */
		if (window_overlap(w->dm->vid1,
				   x_pos(&v), y_pos(&v),
				   v.xres_virtual - v.xres,
				   v.yres_virtual - v.yres))
			RETURN(-EINVAL);
		/* Rule 5 */
		v.bits_per_pixel = 16;
		v.red.offset = 11;
		v.green.offset = 5;
		v.blue.offset = 0;
		v.red.length = 5;
		v.green.length = 6;
		v.blue.length = 5;
		v.transp.offset = v.transp.length = 0;
		v.red.msb_right = v.green.msb_right = v.blue.msb_right
				= v.transp.msb_right = 0;
		v.nonstd = 0;
		v.accel_flags = 0;
	}
	else if (is_win(info->fix.id, OSD1)) {
		v.bits_per_pixel = 4;
	}
	else if (is_win(info->fix.id, VID0)) {
		if (check_new_vid0_size(x_pos(&v), y_pos(&v),
					v.xres_virtual - v.xres,
					v.yres_virtual - v.yres))
			RETURN(-EINVAL);
		v.bits_per_pixel = 16;
	}
	else if (is_win(info->fix.id, VID1)) {
		/* Rule 10 */
		if (window_overlap(w->dm->osd0,
			 	   x_pos(&v), y_pos(&v),
			  	   v.xres_virtual - v.xres,
			  	   v.yres_virtual - v.yres))
			RETURN(-EINVAL);
		v.bits_per_pixel = 16;
	}
	else
		RETURN(-EINVAL);

	memcpy(var, &v, sizeof(v));
	RETURN(0);
}

/* Interlaced = Frame mode, Non-interlaced = Field mode */
static void set_interlaced(char *id, unsigned int on)
{
	on = (on == 0) ? 0 : ~0;

	DBGENTER;
	if (is_win(id, VID0))
		dispc_reg_merge(OSD_VIDWINMD, on, OSD_VIDWINMD_VFF0);
	else if (is_win(id, VID1))
		dispc_reg_merge(OSD_VIDWINMD, on, OSD_VIDWINMD_VFF1);
	else if (is_win(id, OSD0))
		dispc_reg_merge(OSD_OSDWIN0MD, on, OSD_OSDWIN0MD_OFF0);
	else if (is_win(id, OSD1))
		dispc_reg_merge(OSD_OSDWIN1MD, on, OSD_OSDWIN1MD_OFF1);
	DBGEXIT;
}

#if 0 /* unused: commented out  for now */

/* For zooming, we just have to set the start of framebuffer, the zoom factors
 * and the display size. The hardware will then read only
 * (display size / zoom factor) area of the framebuffer and  zoom and
 * display it. In the following function, we assume that the start of
 * framebuffer and the display size parameters are set already.
 */
static void set_zoom(char *id, int h_factor, int v_factor)
{
	DBGENTER;
	if (h_factor != 2 || h_factor != 4 ||
	    v_factor != 2 || v_factor != 4)
		return;

	if (is_win(id, VID0)) {
		dispc_reg_merge(OSD_VIDWINMD,
				h_factor << OSD_VIDWINMD_VHZ0_SHIFT,
				OSD_VIDWINMD_VHZ0);
		dispc_reg_merge(OSD_VIDWINMD,
				v_factor << OSD_VIDWINMD_VVZ0_SHIFT,
				OSD_VIDWINMD_VVZ0);
	}
	else if (is_win(id, VID1)) {
		dispc_reg_merge(OSD_VIDWINMD,
				h_factor << OSD_VIDWINMD_VHZ1_SHIFT,
				OSD_VIDWINMD_VHZ1);
		dispc_reg_merge(OSD_VIDWINMD,
				v_factor << OSD_VIDWINMD_VVZ1_SHIFT,
				OSD_VIDWINMD_VVZ1);
	}
	else if (is_win(id, OSD0)) {
		dispc_reg_merge(OSD_OSDWIN0MD,
				h_factor << OSD_OSDWIN0MD_OHZ0_SHIFT,
				OSD_OSDWIN0MD_OHZ0);
		dispc_reg_merge(OSD_OSDWIN0MD,
				v_factor << OSD_OSDWIN0MD_OVZ0_SHIFT,
				OSD_OSDWIN0MD_OVZ0);
	}
	else if (is_win(id, OSD1)) {
		dispc_reg_merge(OSD_OSDWIN1MD,
				h_factor << OSD_OSDWIN1MD_OHZ1_SHIFT,
				OSD_OSDWIN1MD_OHZ1);
		dispc_reg_merge(OSD_OSDWIN1MD,
				v_factor << OSD_OSDWIN1MD_OVZ1_SHIFT,
				OSD_OSDWIN1MD_OVZ1);
	}
	DBGEXIT;
}
#endif

/* Chooses the ROM CLUT for now. Can be extended later. */
static void set_bg_color(u8 clut, u8 color_offset)
{
	clut = 0;	/* 0 = ROM, 1 = RAM */

	DBGENTER;
	dispc_reg_merge(OSD_MODE, OSD_MODE_BCLUT & clut, OSD_MODE_BCLUT);
	dispc_reg_merge(OSD_MODE, color_offset << OSD_MODE_CABG_SHIFT,
			OSD_MODE_CABG);
	DBGEXIT;
}

static void set_sdram_params(char *id, u32 addr, u32 line_length)
{
	DBGENTER;

	/* The parameters to be written to the registers should be in
	 * multiple of 32 bytes
	 */

	addr = addr;  /* div by 32 */
	line_length = line_length / 32;

	if (is_win(id, VID0)) {
		dispc_reg_out(OSD_VIDWIN0ADR, addr);
		dispc_reg_out(OSD_VIDWIN0OFST, line_length);
		DBGPRINT("Address Calculated =0x%p\n", OSD_VIDWIN0ADR);
		DBGPRINT("Value written =0x%p\n", addr);
		DBGPRINT("****Video Window 0 Address =0x%p\n", dispc_reg_in(OSD_VIDWIN0ADR));
		DBGPRINT("Video Window 0 Line Offset  =0x%p\n", line_length);
	}
	else if (is_win(id, VID1)) {
		dispc_reg_out(OSD_VIDWIN1ADR, addr);
		dispc_reg_out(OSD_VIDWIN1OFST, line_length);
		DBGPRINT("Video Window 1 Address =0x%p\n", addr);
		DBGPRINT("Video Window 1 Line Offset  =0x%p\n", line_length);
	}
	else if (is_win(id, OSD0)) {
		dispc_reg_out(OSD_OSDWIN0ADR, addr);
		dispc_reg_out(OSD_OSDWIN0OFST, line_length);
		DBGPRINT("OSD Window 0 Address =0x%p\n", addr);
		DBGPRINT("OSD Window 0 Line Offset  =0x%p\n", line_length);
	}
	else if (is_win(id, OSD1)) {
		dispc_reg_out(OSD_OSDWIN1ADR, addr);
		dispc_reg_out(OSD_OSDWIN1OFST, line_length);
		DBGPRINT("OSD Window 1 Address =0x%p\n", addr);
		DBGPRINT("OSD Window 1 Line Offset  =0x%p\n", line_length);
	}

	DBGEXIT;
}

static void set_win_enable(char *id, unsigned int on)
{
	on = (on == 0) ? 0 : ~0;

	DBGENTER;
	if (is_win(id, VID0))
		dispc_reg_merge(OSD_VIDWINMD, on, OSD_VIDWINMD_ACT0);
	else if (is_win(id, VID1))
		dispc_reg_merge(OSD_VIDWINMD, on, OSD_VIDWINMD_ACT1);
	else if (is_win(id, OSD0))
		dispc_reg_merge(OSD_OSDWIN0MD, on, OSD_OSDWIN0MD_OACT0);
	else if (is_win(id, OSD1)) {
		/* The OACT1 bit is applicable only if OSD1 is not used as
		 * the attribute window
		 */
		if (!(dispc_reg_in(OSD_OSDWIN1MD) & OSD_OSDWIN1MD_OASW))
			dispc_reg_merge(OSD_OSDWIN1MD, on, OSD_OSDWIN1MD_OACT1);
	}
	DBGEXIT;
}

static void set_win_mode(char *id)
{
	DBGENTER;
	if (is_win(id, VID0))
		;
	else if (is_win(id, VID1))
		;
	else if (is_win(id, OSD0))
		/* Set RGB565 mode */
		dispc_reg_merge(OSD_OSDWIN0MD, OSD_OSDWIN0MD_RGB0E,
			OSD_OSDWIN0MD_RGB0E);
	else if (is_win(id, OSD1)) {
		/* Set as attribute window */
		dispc_reg_merge(OSD_OSDWIN1MD, OSD_OSDWIN1MD_OASW,
			OSD_OSDWIN1MD_OASW);
	}
	DBGEXIT;

}


/**
 *      dm64xxfb_set_par - Optional function. Alters the hardware state.
 *      @info: frame buffer structure that represents a single frame buffer
 *
 *	Using the fb_var_screeninfo in fb_info we set the resolution of the
 *	this particular framebuffer. This function alters the par AND the
 *	fb_fix_screeninfo stored in fb_info. It doesn't not alter var in
 *	fb_info since we are using that data. This means we depend on the
 *	data in var inside fb_info to be supported by the hardware.
 *	dm64xxfb_check_var is always called before dmfb_set_par to ensure this.
 *	Again if you can't can't the resolution you don't need this function.
 *
 */
static int dm64xxfb_set_par(struct fb_info *info)
{
	struct dm_win_info *w = (struct dm_win_info *) info->par;
	struct fb_var_screeninfo *v = &info->var;
	u32 start = 0, offset = 0;

	DBGENTER;

	DBGPRINT(" set_par for %s \n", info->fix.id);
	info->fix.line_length = v->xres_virtual * v->bits_per_pixel / 8;

	offset = v->yoffset * info->fix.line_length +
		 v->xoffset * v->bits_per_pixel / 8;
	start = (u32) w->fb_base_phys + offset;
	set_sdram_params(info->fix.id, start, info->fix.line_length);

	set_interlaced(info->fix.id, 1);
	set_win_position(info->fix.id,
			 x_pos(v),y_pos(v), v->xres, v->yres / 2);
	set_win_mode(info->fix.id);
	set_win_enable(info->fix.id, 1);
	venc_start();

	RETURN(0);
}

/**
 *	dm64xxfb_ioctl - handler for private ioctls.
 */
static int
dm64xxfb_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	unsigned long arg, struct fb_info *info)
{
	struct dm_win_info *w = (struct dm_win_info *) info->par;

	DBGENTER;
	switch (cmd) {
		case FBIO_WAITFORVSYNC:
			/* This ioctl accepts an integer argument to specify a
			 * display.  We only support one display, so we will
			 * simply ignore the argument.
			 */
			RETURN(dm64xxfb_wait_for_vsync(w));
	}
	RETURN(-EINVAL);
}

/**
 *  	dm64xxfb_setcolreg - Optional function. Sets a color register.
 *      @regno: Which register in the CLUT we are programming
 *      @red: The red value which can be up to 16 bits wide
 *	@green: The green value which can be up to 16 bits wide
 *	@blue:  The blue value which can be up to 16 bits wide.
 *	@transp: If supported the alpha value which can be up to 16 bits wide.
 *      @info: frame buffer info structure
 *
 *  	Set a single color register. The values supplied have a 16 bit
 *  	magnitude which needs to be scaled in this function for the hardware.
 *	Things to take into consideration are how many color registers, if
 *	any, are supported with the current color visual. With truecolor mode
 *	no color palettes are supported. Here a psuedo palette is created
 *	which we store the value in pseudo_palette in struct fb_info. For
 *	pseudocolor mode we have a limited color palette. To deal with this
 *	we can program what color is displayed for a particular pixel value.
 *	DirectColor is similar in that we can program each color field. If
 *	we have a static colormap we don't need to implement this function.
 *
 *	Returns negative errno on error, or zero on success.
 */
static int dm64xxfb_setcolreg(unsigned regno, unsigned red, unsigned green,
			      unsigned blue, unsigned transp,
			      struct fb_info *info)
{
	/* only pseudo-palette (16 bpp) allowed */
	if (regno >= 16)  /* maximum number of palette entries */
		RETURN(1);

	if (info->var.grayscale) {
		/* grayscale = 0.30*R + 0.59*G + 0.11*B */
		red = green = blue = (red * 77 + green * 151 + blue * 28) >> 8;
	}

	/* Truecolor has hardware-independent 16-entry pseudo-palette */
	if (info->fix.visual == FB_VISUAL_TRUECOLOR) {
		u32 v;

		if (regno >= 16)
			RETURN(1);

		red >>= (16 - info->var.red.length);
		green >>= (16 - info->var.green.length);
		blue >>= (16 - info->var.blue.length);

		v = (red << info->var.red.offset) |
		    (green << info->var.green.offset) |
		    (blue << info->var.blue.offset);

		switch (info->var.bits_per_pixel) {
			case 16:
		   		((u16*)(info->pseudo_palette))[regno] = v;
				break;
			default:
				RETURN(1);
		}
		RETURN(0);
	}
	RETURN(0);
}

/**
 *      dm64xxfb_pan_display - NOT a required function. Pans the display.
 *      @var: frame buffer variable screen structure
 *      @info: frame buffer structure that represents a single frame buffer
 *
 *	Pan (or wrap, depending on the `vmode' field) the display using the
 *  	`xoffset' and `yoffset' fields of the `var' structure.
 *  	If the values don't fit, return -EINVAL.
 *
 *      Returns negative errno on error, or zero on success.
 */
static int dm64xxfb_pan_display(struct fb_var_screeninfo *var,
				struct fb_info *info)
{
	struct dm_win_info *w = (struct dm_win_info *) info->par;
	u32 start = 0, offset = 0;

	DBGENTER;
	if (var->xoffset > var->xres_virtual - var->xres)
		RETURN(-EINVAL);
	if (var->yoffset > var->yres_virtual - var->yres)
		RETURN(-EINVAL);
	if((var->xres_virtual * var->bits_per_pixel/8) % 32)
		RETURN(-EINVAL);

	offset = var->yoffset * info->fix.line_length +
		 var->xoffset * var->bits_per_pixel / 8;
	start = (u32) w->fb_base_phys + offset;
	set_sdram_params(info->fix.id, start, info->fix.line_length);

	RETURN(0);
}

/**
 *      dm64xxfb_blank - NOT a required function. Blanks the display.
 *      @blank_mode: the blank mode we want.
 *      @info: frame buffer structure that represents a single frame buffer
 *
 *      Blank the screen if blank_mode != 0, else unblank. Return 0 if
 *      blanking succeeded, != 0 if un-/blanking failed due to e.g. a
 *      video mode which doesn't support it. Implements VESA suspend
 *      and powerdown modes on hardware that supports disabling hsync/vsync:
 *      blank_mode == 2: suspend vsync
 *      blank_mode == 3: suspend hsync
 *      blank_mode == 4: powerdown
 *
 *      Returns negative errno on error, or zero on success.
 *
 */
static int dm64xxfb_blank(int blank_mode, struct fb_info *info)
{
	DBGENTER;
	/* ... */
	RETURN(0);
}

/**
 *	dm64xxfb_sync - NOT a required function. Normally the accel engine
 *		     for a graphics card take a specific amount of time.
 *		     Often we have to wait for the accelerator to finish
 *		     its operation before we can write to the framebuffer
 *		     so we can have consistent display output.
 *
 *      @info: frame buffer structure that represents a single frame buffer
 */
int dm64xxfb_sync(struct fb_info *info)
{
//	DBGENTER;
//	RETURN(0);
	return 0;
}

    /*
     *  Setup
     */

/*
 * Only necessary if your driver takes special options,
 * otherwise we fall back on the generic fb_setup().
 */
int __init dm64xxfb_setup(char *options)
{
	DBGENTER;
	/* Parse user specified options (`video=dm64xxfb:') */
	RETURN(0);
}

static int mem_release(struct dm_win_info *w)
{
	DBGENTER;
	if (!w->alloc_fb_mem) {
		iounmap((void*)w->fb_base);
		release_mem_region(w->fb_base_phys, w->fb_size);
	}
	else
		dma_free_coherent(NULL, w->fb_size, (void *)w->fb_base,
				  w->fb_base_phys);
	kfree(w);
	RETURN(0);
}

static int mem_alloc(struct dm_win_info **win, dma_addr_t fb_base_phys,
		     unsigned long fb_size, char *fbname)
{
	struct dm_win_info *w;

	DBGENTER;
	w = kmalloc(sizeof(struct dm_win_info), GFP_KERNEL);
	if (!w) {
		printk(KERN_ERR "%s: could not allocate memory\n", fbname);
		RETURN(-ENOMEM);
	}
	memset(w, 0, sizeof(struct dm_win_info));

	w->fb_base_phys = fb_base_phys;
	w->fb_size = fb_size;

	/* A null base address indicates that the framebuffer memory should be
	 * dynamically allocated.
	 */
	if (!w->fb_base_phys)
		w->alloc_fb_mem = 1;

	if (!w->alloc_fb_mem) {
		if (!request_mem_region(w->fb_base_phys, w->fb_size, fbname))
		{
			printk(KERN_ERR
				"%s: cannot reserve FB region\n", fbname);
			goto free_par;
		}
		w->fb_base =
			(unsigned long) ioremap(w->fb_base_phys, w->fb_size);
		if (!w->fb_base) {
			printk(KERN_ERR "%s: cannot map framebuffer\n", fbname);
			goto release_fb;
		}
	}
	else {
		/* allocate coherent memory for the framebuffer */
		w->fb_base = (unsigned long) dma_alloc_coherent(NULL,
			w->fb_size, &w->fb_base_phys, GFP_KERNEL | GFP_DMA);
		if (!w->fb_base) {
			printk(KERN_ERR "%s: cannot allocate "
				"framebuffer\n", fbname);
			goto free_par;
		}
		DBGPRINT("Framebuffer allocated at 0x%lx, mapped to 0x%p, size %dk\n",
	       w->fb_base_phys, w->fb_base, w->fb_size/1024);
	}

	*win = w;
	RETURN(0);

release_fb:
	if (!w->alloc_fb_mem)
		release_mem_region(w->fb_base_phys, w->fb_size);
free_par:
	kfree(w);
	RETURN(-ENOMEM);
}

static struct fb_info *init_fb_info(struct dm_win_info *w,
				    struct fb_var_screeninfo *var, char *id)
{
	struct fb_info *info = &(w->info);
	struct dm_info *dm = w->dm;

	DBGENTER;

	DBGPRINT("Entered the Davinci FB Init Routine\n");

	/* initialize the fb_info structure */
	info->flags = FBINFO_DEFAULT;
	info->fbops = &dm64xxfb_ops;
	info->screen_base = (char *)(w->fb_base);
/* 	info->currcon = -1; */
	info->pseudo_palette = w->pseudo_palette;
	info->par = w;

	/* Initialize variable screeninfo.
	 * The variable screeninfo can be directly specified by the user
	 * via an ioctl.
	 */
	memcpy(&info->var, var, sizeof(info->var));
	info->var.activate = FB_ACTIVATE_NOW;

	/* Initialize fixed screeninfo.
	 * The fixed screeninfo cannot be directly specified by the user, but
	 * it may change to reflect changes to the var info.
	 */
	strlcpy(info->fix.id, id, sizeof(info->fix.id));
	info->fix.smem_start = w->fb_base_phys;
	info->fix.line_length =
			(info->var.xres_virtual*info->var.bits_per_pixel)/8;
	info->fix.smem_len = w->fb_size;
	info->fix.type = FB_TYPE_PACKED_PIXELS;
	info->fix.visual = (info->var.bits_per_pixel <= 8) ?
				FB_VISUAL_PSEUDOCOLOR: FB_VISUAL_TRUECOLOR;
	info->fix.xpanstep = 0;
	info->fix.ypanstep = 0;
	info->fix.ywrapstep = 0;
	info->fix.type_aux = 0;
	info->fix.mmio_start = dm->mmio_base_phys;
	info->fix.mmio_len = dm->mmio_size;
	info->fix.accel = FB_ACCEL_NONE;

	DBGEXIT;
	return info;
}

static void venc_config(void)
{
	DBGENTER;
	/* start by resetting video encoder module */
	dispc_reg_out(VENC_VMOD, 0);

	/* Enable Composite output and start video encoder */
	dispc_reg_out(VENC_VMOD, (VENC_VMOD_VIE | VENC_VMOD_VENC));
	DBGEXIT;
}

static void venc_start(void)
{
	DBGENTER;
	/* Set REC656 Mode */
	dispc_reg_out(VENC_YCCCTL, 0x1);

	/* Enable output mode and NTSC  */
	dispc_reg_out(VENC_VMOD, 0x1003);

	/* Enable all DACs  */
	dispc_reg_out(VENC_DACTST, 0);

	DBGEXIT;
}

/*
static void show_regs()
{
	u32 off = OSD_MODE; // DM342_OSDMODE
	int i = 0;

	for

}
*/

static void dm64xxfb_cleanup(void);
    /*
     *  Initialization
     */

int __init dm64xxfb_init(void)
{
	struct fb_info *info;

	DBGENTER;

	dm.mmio_base_phys = OSD_REG_BASE;
	dm.mmio_size = OSD_REG_SIZE;

	/* request the mem regions */
#ifdef REVISIT
	if (!request_mem_region(dm.mmio_base_phys, dm.mmio_size, DRIVER)) {
		printk(KERN_ERR DRIVER ": cannot reserve MMIO region\n");
		RETURN(-ENODEV);
	}

	/* map the regions */
	dm.mmio_base =
		(unsigned long) ioremap(dm.mmio_base_phys, dm.mmio_size);
	if (!dm.mmio_base) {
		printk(KERN_ERR DRIVER ": cannot map MMIO\n");
		goto release_mmio;
	}
#endif
	/* initialize the vsync wait queue */
	init_waitqueue_head(&dm.vsync_wait);
	dm.timeout = HZ/5;

	DBGPRINT("Setting Up Clocks for DM420 OSD\n");

	/* Initialize the VPSS Clock Control register */
	dispc_reg_out(VPSS_CLKCTL, 0x18);

	/* Set Base Pixel X and Base Pixel Y */
	dispc_reg_out(OSD_BASEPX, BASEX);

	dispc_reg_out(OSD_BASEPY, BASEY);

	/* Reset OSD registers to default. */
	dispc_reg_out(OSD_MODE, 0);
	dispc_reg_out(OSD_OSDWIN0MD, 0);
//REVISIT:  should we reset the other window modes too?

	/* Set blue background color */
	set_bg_color(0, 162);

	/* OSD0 Distortion Fix */
	dispc_reg_out(OSD_MODE, 0x200);

	venc_config();

	DBGPRINT("-------------- VID0 -----------------\n");

	/* Setup VID0 framebuffer */
	if (!mem_alloc(&dm.vid0, VID0_FB_PHY, VID0_FB_SIZE, VID0_FBNAME)) {
		dm.vid0->dm = &dm;
		info = init_fb_info(dm.vid0, &vid0_default_var, VID0_FBNAME);
		if (dm64xxfb_check_var(&info->var, info)) {
			printk(KERN_ERR VID0_FBNAME
				": invalid default video mode\n");
			goto exit;
		}
		memset((void *)dm.vid0->fb_base, 0x88, dm.vid0->fb_size);
	} else
		goto exit;

	DBGPRINT("-------------- OSD0 -----------------\n");
	/* Setup OSD0 framebuffer */
	if (!mem_alloc(&dm.osd0, OSD0_FB_PHY, OSD0_FB_SIZE, OSD0_FBNAME)) {
		dm.osd0->dm = &dm;
		info = init_fb_info(dm.osd0, &osd0_default_var, OSD0_FBNAME);
		if (dm64xxfb_check_var(&info->var, info)) {
			printk(KERN_ERR OSD0_FBNAME
				": invalid default video mode\n");
			goto exit_osd0;
		}
		memset((void *)dm.osd0->fb_base, 0, dm.osd0->fb_size);
	}

	DBGPRINT("-------------- OSD1 -----------------\n");
	/* Setup OSD1 framebuffer */
	if (!mem_alloc(&dm.osd1, OSD1_FB_PHY, OSD1_FB_SIZE, OSD1_FBNAME)) {
		dm.osd1->dm = &dm;
		info = init_fb_info(dm.osd1, &osd1_default_var, OSD1_FBNAME);
		if (dm64xxfb_check_var(&info->var, info)) {
			printk(KERN_ERR OSD1_FBNAME
				": invalid default video mode\n");
			goto exit_osd1;
		}
		/* Set blend factor to show OSD windows */
		memset((void *)dm.osd1->fb_base, 0xff, dm.osd1->fb_size);
	}

	DBGPRINT("-------------- VID1 -----------------\n");
	/* Setup VID1 framebuffer */
	if (!mem_alloc(&dm.vid1, VID1_FB_PHY, VID1_FB_SIZE, VID1_FBNAME)) {
		dm.vid1->dm = &dm;
		info = init_fb_info(dm.vid1, &vid1_default_var, VID1_FBNAME);
		if (dm64xxfb_check_var(&info->var, info)) {
			printk(KERN_ERR VID1_FBNAME
				": invalid default video mode\n");
			goto exit_vid1;
		}
		memset((void *)dm.vid1->fb_base, 8, dm.vid1->fb_size);
	}

	/* Register OSD0 framebuffer */
	if (dm.osd0) {
		info = &dm.osd0->info;
		if (register_framebuffer(info) < 0) {
			printk(KERN_ERR OSD0_FBNAME
				"Unable to register OSD0 framebuffer\n");
exit_osd0:		mem_release(dm.osd0);
		} else
			printk(KERN_INFO "fb%d: %s frame buffer device\n",
				info->node, info->fix.id);
	}

	/* Register VID0 framebuffer */
	info = &dm.vid0->info;
	if (register_framebuffer(info) < 0) {
		printk(KERN_ERR VID0_FBNAME
			"Unable to register VID0 framebuffer\n");
		goto exit;
	} else {
		printk(KERN_INFO "fb%d: %s frame buffer device\n",
			info->node, info->fix.id);
		dm64xxfb_set_par(info);
	}

	/* Register OSD1 framebuffer */
	if (dm.osd1) {
		info = &dm.osd1->info;
		if (register_framebuffer(info) < 0) {
			printk(KERN_ERR OSD1_FBNAME
				"Unable to register OSD1 framebuffer\n");
exit_osd1:		mem_release(dm.osd1);
		} else {
			DBGPRINT(KERN_INFO "fb%d: %s frame buffer device\n",
				info->node, info->fix.id);
			dm64xxfb_set_par(info);
		}
	}

	/* Register VID1 framebuffer */
	if (dm.vid1) {
		info = &dm.vid1->info;
		if (register_framebuffer(info) < 0) {
			mem_release(dm.vid1);
			printk(KERN_ERR VID1_FBNAME
				"Unable to register VID1 framebuffer\n");
exit_vid1:		mem_release(dm.vid1);
		} else {
			DBGPRINT(KERN_INFO "fb%d: %s frame buffer device\n",
				info->node, info->fix.id);
			dm64xxfb_set_par(info);
		}
	}

        /* install our interrupt service routine */
	if (request_irq(IRQ_VENCINT, dm64xxfb_isr, SA_SHIRQ, DRIVER, &dm))
	{
		printk(KERN_ERR DRIVER
			": could not install interrupt service routine\n");
		goto exit;
	}
	RETURN(0);

exit:
	dm64xxfb_cleanup();
	RETURN(-ENODEV);

	iounmap((void*)dm.mmio_base);

	release_mem_region(dm.mmio_base_phys, dm.mmio_size);
	RETURN(-ENODEV);
}
    /*
     *  Cleanup
     */

static void dm64xxfb_cleanup(void)
{
	DBGENTER;

	free_irq(IRQ_VENCINT, &dm);

	/* Cleanup all framebuffers */
	if (dm.osd0) {
		unregister_framebuffer(&dm.osd0->info);
		mem_release(dm.osd0);
	}
	if (dm.osd1) {
		unregister_framebuffer(&dm.osd1->info);
		mem_release(dm.osd1);
	}
	if (dm.vid0) {
		unregister_framebuffer(&dm.vid0->info);
		mem_release(dm.vid0);
	}
	if (dm.vid1) {
		unregister_framebuffer(&dm.vid1->info);
		mem_release(dm.vid1);
	}

	if(dm.mmio_base)
		iounmap((void*)dm.mmio_base);
	release_mem_region(dm.mmio_base_phys, dm.mmio_size);

	DBGEXIT;
}

/* ------------------------------------------------------------------------- */

    /*
     *  Frame buffer operations
     */
static struct fb_ops dm64xxfb_ops = {
	.owner		= THIS_MODULE,
	.fb_check_var	= dm64xxfb_check_var,
	.fb_set_par	= dm64xxfb_set_par,
	.fb_setcolreg	= dm64xxfb_setcolreg,
	.fb_blank	= dm64xxfb_blank,
	.fb_pan_display	= dm64xxfb_pan_display,
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
	.fb_cursor	= soft_cursor,
	.fb_rotate	= NULL,
	.fb_sync	= dm64xxfb_sync,
	.fb_ioctl	= dm64xxfb_ioctl,
};

/* ------------------------------------------------------------------------- */

    /*
     *  Modularization
     */

module_init(dm64xxfb_init);
module_exit(dm64xxfb_cleanup);

MODULE_DESCRIPTION("Framebuffer driver for TI DM644x DV-EVM Board");
MODULE_AUTHOR("Texas Instruments");
MODULE_LICENSE("GPL");

