/*
 * Unix SMB/CIFS implementation. 
 * Copyright (C) Jeremy Allison 1995-1998
 * Copyright (C) Tim Potter     2001
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675
 * Mass Ave, Cambridge, MA 02139, USA.  */

#include "includes.h"



static char *passwd_filename = 0;


void _fDebug(char *fmt, ...)
{
}

void E_md4hash(const char *passwd, uchar p16[16])
{
	int len;
	smb_ucs2_t wpwd[129];
	int i;

	
	/* Password must be converted to NT unicode - null terminated. */
	len = strlen(passwd);
#if 0
	push_ucs2(NULL, wpwd, (const char *)passwd, 256, STR_UNICODE|STR_NOALIGN|STR_TERMINATE);
#else
	for (i = 0; i < len; i++) {
		wpwd[i] = (unsigned char)passwd[i];
	}
	wpwd[i] = 0; // termination
#endif
	/* Calculate length in bytes */
	len = len /*strlen_w(wpwd)*/ * sizeof(int16);

	mdfour(p16, (unsigned char *)wpwd, len);
	ZERO_STRUCT(wpwd);	
}

/**
 * Creates the DES forward-only Hash of the users password in DOS ASCII charset
 * @param passwd password in 'unix' charset.
 * @param p16 return password hashed with DES, caller allocated 16 byte buffer
 * @return False if password was > 14 characters, and therefore may be incorrect, otherwise True
 * @note p16 is filled in regardless
 */
 
BOOL E_deshash(const char *passwd, uchar p16[16])
{
	BOOL ret = True;
	char dospwd[256+2];
	int i;
	int len;
	
	/* Password must be converted to DOS charset - null terminated, uppercase. */
//	push_ascii(dospwd, passwd, sizeof(dospwd), STR_UPPER|STR_TERMINATE);
	len = strlen(passwd);
	for (i = 0; i < len; i++) {
		char c = passwd[i];
		if (islower(c)) c = toupper(c);
		dospwd[i] = c;
	}
	dospwd[i] = 0;
       
	/* Only the fisrt 14 chars are considered, password need not be null terminated. */
	E_P16((const unsigned char *)dospwd, p16);

	if (strlen(dospwd) > 14) {
		ret = False;
	}

	memset(dospwd, 0, sizeof(dospwd));
	// ZERO_STRUCT(dospwd);	

	return ret;
}

static void my_pdb_sethexpwd(char *p, const unsigned char *pwd)
{
	if (pwd != NULL) {
		int i;
		for (i = 0; i < 16; i++)
			slprintf(&p[i*2], 3, "%02X", pwd[i]);
	} else {
		strncpy(p, "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX", 32);
	}
}

static BOOL my_mod_smbfilepwd_entry(char *filename, 
								char *username,
								char *smb_passwd,
								char *smb_nt_passwd)
{
	/* Static buffers we will return. */
	pstring user_name;

	char linebuf[256];
	char readbuf[1024];
	unsigned char c;
	fstring ascii_p16;
	fstring encode_bits;
	unsigned char *p = NULL;
	size_t linebuf_len = 0;
	FILE *fp;
	BOOL found_entry = False;
	BOOL got_pass_last_set_time = False;

	SMB_OFF_T pwd_seekpos = 0;

	int i;
	int wr_len;
	int fd;


	fp = fopen(filename, "r+");

	if (fp == NULL) {
		return False;
	}
	/* Set a buffer to do more efficient reads */
	setvbuf(fp, readbuf, _IOFBF, sizeof(readbuf));

	/* Make sure it is only rw by the owner */
	chmod(filename, 0600);

	/* We have a write lock on the file. */
	/*
	 * Scan the file, a line at a time and check if the name matches.
	 */
	while (!feof(fp)) {
		pwd_seekpos = ftell(fp);

		linebuf[0] = '\0';

		fgets(linebuf, sizeof(linebuf), fp);
		if (ferror(fp)) {
			fclose(fp);
			return False;
		}

		/*
		 * Check if the string is terminated with a newline - if not
		 * then we must keep reading and discard until we get one.
		 */
		linebuf_len = strlen(linebuf);
		if (linebuf[linebuf_len - 1] != '\n') {
			c = '\0';
			while (!ferror(fp) && !feof(fp)) {
				c = fgetc(fp);
				if (c == '\n') {
					break;
				}
			}
		} else {
			linebuf[linebuf_len - 1] = '\0';
		}

#ifdef DEBUG_PASSWORD
		DEBUG(100, ("mod_smbfilepwd_entry: got line |%s|\n", linebuf));
#endif

		if ((linebuf[0] == 0) && feof(fp)) {
			DEBUG(4, ("mod_smbfilepwd_entry: end of file reached\n"));
			break;
		}

		/*
		 * The line we have should be of the form :-
		 * 
		 * username:uid:[32hex bytes]:....other flags presently
		 * ignored....
		 * 
		 * or,
		 *
		 * username:uid:[32hex bytes]:[32hex bytes]:[attributes]:LCT-XXXXXXXX:...ignored.
		 *
		 * if Windows NT compatible passwords are also present.
		 */

		if (linebuf[0] == '#' || linebuf[0] == '\0') {
			DEBUG(6, ("mod_smbfilepwd_entry: skipping comment or blank line\n"));
			continue;
		}

		// p = (unsigned char *) strchr_m(linebuf, ':');
		p = (unsigned char *) strchr(linebuf, ':');

		if (p == NULL) {
			DEBUG(0, ("mod_smbfilepwd_entry: malformed password entry (no :)\n"));
			continue;
		}

		/*
		 * As 256 is shorter than a pstring we don't need to check
		 * length here - if this ever changes....
		 */

		SMB_ASSERT(sizeof(user_name) > sizeof(linebuf));

		strncpy(user_name, linebuf, PTR_DIFF(p, linebuf));
		user_name[PTR_DIFF(p, linebuf)] = '\0';
		if (!strcmp(user_name, username)) {
			found_entry = True;
			break;
		}
	}

	if (!found_entry) {
		fclose(fp);

		DEBUG(2, ("Cannot update entry for user %s, as they don't exist in the smbpasswd file!\n",
			username));
		return False;
	}

	DEBUG(6, ("mod_smbfilepwd_entry: entry exists for user %s\n", username));

	/* User name matches - get uid and password */
	p++; /* Go past ':' */

	if (!isdigit(*p)) {
		DEBUG(0, ("mod_smbfilepwd_entry: malformed password entry for user %s (uid not number)\n",
			username));
		fclose(fp);
		return False;
	}

	while (*p && isdigit(*p)) {
		p++;
	}
	if (*p != ':') {
		DEBUG(0, ("mod_smbfilepwd_entry: malformed password entry for user %s (no : after uid)\n",
			username));
		fclose(fp);
		return False;
	}

	/*
	 * Now get the password value - this should be 32 hex digits
	 * which are the ascii representations of a 16 byte string.
	 * Get two at a time and put them into the password.
	 */
	p++;

	/* Record exact password position */
	pwd_seekpos += PTR_DIFF(p, linebuf);

	if (linebuf_len < (PTR_DIFF(p, linebuf) + 33)) {
		DEBUG(0, ("mod_smbfilepwd_entry: malformed password entry for user %s (passwd too short)\n",
			username));
		fclose(fp);
		return (False);
	}

	if (p[32] != ':') {
		DEBUG(0, ("mod_smbfilepwd_entry: malformed password entry for user %s (no terminating :)\n",
			username));
		fclose(fp);
		return False;
	}

	/* Now check if the NT compatible password is available. */
	p += 33; /* Move to the first character of the line after the lanman password. */
	if (linebuf_len < (PTR_DIFF(p, linebuf) + 33)) {
		DEBUG(0, ("mod_smbfilepwd_entry: malformed password entry for user %s (passwd too short)\n",
			username));
		fclose(fp);
		return (False);
	}

	if (p[32] != ':') {
		DEBUG(0, ("mod_smbfilepwd_entry: malformed password entry for user %s (no terminating :)\n",
			username));
		fclose(fp);
		return False;
	}

	/* 
	 * Now check if the account info and the password last
	 * change time is available.
	 */
	p += 33; /* Move to the first character of the line after the NT password. */

	if (*p == '[') {
		i = 0;
		encode_bits[i++] = *p++;
		while((linebuf_len > PTR_DIFF(p, linebuf)) && (*p != ']')) {
			encode_bits[i++] = *p++;
		}

		encode_bits[i++] = ']';
		encode_bits[i++] = '\0';

		if(i == NEW_PW_FORMAT_SPACE_PADDED_LEN) {
			/*
			 * We are using a new format, space padded
			 * acct ctrl field. Encode the given acct ctrl
			 * bits into it.
			 */
// AVM 	fstrcpy(encode_bits, pdb_encode_acct_ctrl(pwd->acct_ctrl, NEW_PW_FORMAT_SPACE_PADDED_LEN));
		} else {
			fclose(fp);
			return False;
		}

		/* Go past the ']' */
		if(linebuf_len > PTR_DIFF(p, linebuf)) {
			p++;
		}

		if((linebuf_len > PTR_DIFF(p, linebuf)) && (*p == ':')) {
			p++;

			/* We should be pointing at the LCT entry. */
			// if((linebuf_len > (PTR_DIFF(p, linebuf) + 13)) && (StrnCaseCmp((char *)p, "LCT-", 4) == 0)) {
			if((linebuf_len > (PTR_DIFF(p, linebuf) + 13)) && (strncmp((char *)p, "LCT-", 4) == 0)) {
				p += 4;
				for(i = 0; i < 8; i++) {
					if(p[i] == '\0' || !isxdigit(p[i])) {
						break;
					}
				}
				if(i == 8) {
					/*
					 * p points at 8 characters of hex digits -
					 * read into a time_t as the seconds since
					 * 1970 that the password was last changed.
					 */
					got_pass_last_set_time = True;
				} /* i == 8 */
			} /* *p && StrnCaseCmp() */
		} /* p == ':' */
	} /* p == '[' */

	/* Entry is correctly formed. */

	/* Create the 32 byte representation of the new p16 */
	my_pdb_sethexpwd(ascii_p16, smb_passwd);

	/* Add on the NT md4 hash */
	ascii_p16[32] = ':';
	wr_len = 66;
	my_pdb_sethexpwd(ascii_p16+33, smb_nt_passwd);
	ascii_p16[65] = ':';
	ascii_p16[66] = '\0'; /* null-terminate the string so that strlen works */

#if 0 /* AVM */
	/* Add on the account info bits and the time of last password change. */
	if(got_pass_last_set_time) {
		slprintf(&ascii_p16[strlen(ascii_p16)], 
			sizeof(ascii_p16)-(strlen(ascii_p16)+1),
			"%s:LCT-%08X:", 
			encode_bits, (uint32)pwd->pass_last_set_time );
		wr_len = strlen(ascii_p16);
	}
#endif

#ifdef DEBUG_PASSWORD
	DEBUG(100,("mod_smbfilepwd_entry: "));
	dump_data(100, ascii_p16, wr_len);
#endif

	if(wr_len > sizeof(linebuf)) {
		DEBUG(0, ("mod_smbfilepwd_entry: line to write (%d) is too long.\n", wr_len+1));
		fclose(fp);
		return (False);
	}

	/*
	 * Do an atomic write into the file at the position defined by
	 * seekpos.
	 */

	/* The mod user write needs to be atomic - so get the fd from 
		the fp and do a raw write() call.
	 */

	fd = fileno(fp);

	if (lseek(fd, pwd_seekpos - 1, SEEK_SET) != pwd_seekpos - 1) {
		fclose(fp);
		return False;
	}

	/* Sanity check - ensure the areas we are writing are framed by ':' */
	if (read(fd, linebuf, wr_len+1) != wr_len+1) {
		fclose(fp);
		return False;
	}

	if ((linebuf[0] != ':') || (linebuf[wr_len] != ':'))	{
		fclose(fp);
		return False;
	}
 
	if (lseek(fd, pwd_seekpos, SEEK_SET) != pwd_seekpos) {
		fclose(fp);
		return False;
	}

	if (write(fd, ascii_p16, wr_len) != wr_len) {
		fclose(fp);
		return False;
	}

	fclose(fp);
	return True;
}

static int my_password_change(const char *user_name,
			   const char *new_passwd)
{
	uchar new_lanman_p16[LM_HASH_LEN];
	uchar new_nt_p16[NT_HASH_LEN];
	int ret;

	/* Calculate the MD4 hash (NT compatible) of the password */
	E_md4hash(new_passwd, new_nt_p16);

	if (!E_deshash(new_passwd, new_lanman_p16)) {
		/* E_deshash returns false for 'long' passwords (> 14
		   DOS chars).  This allows us to match Win2k, which
		   does not store a LM hash for these passwords (which
		   would reduce the effective password length to 14 */

		memset(new_lanman_p16, 0, sizeof(new_lanman_p16));
	}

	 ret = my_mod_smbfilepwd_entry(passwd_filename, user_name,
								new_lanman_p16,
								new_nt_p16);

	 return ret;
}


static const char g_username[] = "ftpuser";

/*********************************************************
 Start here.
**********************************************************/
int main(int argc, char **argv)
{	
	int  ret;

	passwd_filename = "/var/samba/private/smbpasswd";

	if (argc == 2) {
		ret = my_password_change(g_username, argv[1]);
	} else {
		fprintf(stderr, "use: smbpasswd <password>\n");
		fprintf(stderr, "       <password> will become the new samba password for user %s in %s\n", g_username, passwd_filename);
		return -9;
	}

	if (ret) {
		// ok
fprintf(stderr, "samba password for user %s changed to %s in %s\n", g_username, argv[1], passwd_filename);
		return 0;
	} 
fprintf(stderr, "setting of new samba password for user %s in %s FAILED\n", g_username, passwd_filename);
	return -1;
}
